# © Copyright Databand.ai, an IBM Company 2022

import atexit
import logging
import os
import sys
import typing

from subprocess import list2cmdline
from typing import Optional

import dbnd

from dbnd._core.configuration import get_dbnd_project_config
from dbnd._core.configuration.config_value import ConfigValuePriority
from dbnd._core.configuration.dbnd_config import config
from dbnd._core.configuration.environ_config import try_get_script_name
from dbnd._core.constants import RunState, TaskRunState, UpdateSource
from dbnd._core.context.databand_context import new_dbnd_context
from dbnd._core.current import try_get_databand_run
from dbnd._core.log.config import FORMAT_SIMPLE
from dbnd._core.parameter.parameter_value import Parameters
from dbnd._core.run.databand_run import new_databand_run
from dbnd._core.settings import TrackingConfig
from dbnd._core.task.tracking_task import TrackingTask
from dbnd._core.task_build.task_definition import TaskDefinition
from dbnd._core.task_build.task_passport import TaskPassport
from dbnd._core.task_build.task_source_code import TaskSourceCode
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.task_run.task_run_error import TaskRunError
from dbnd._core.tracking.airflow_dag_inplace_tracking import (
    build_run_time_airflow_task,
    override_airflow_log_system_for_tracking,
)
from dbnd._core.tracking.managers.callable_tracking import _handle_tracking_error
from dbnd._core.tracking.schemas.tracking_info_run import RootRunInfo
from dbnd._core.utils import seven
from dbnd._core.utils.airflow_utils import get_project_name_from_airflow_tags
from dbnd._core.utils.timezone import utcnow
from dbnd._core.utils.uid_utils import get_job_run_uid, get_task_run_uid
from dbnd._vendor import pendulum


logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from dbnd._core.context.databand_context import DatabandContext
    from dbnd._core.run.databand_run import DatabandRun

    T = typing.TypeVar("T")


# Helper functions


def _set_tracking_config_overide(airflow_context=None):
    # Ceate proper DatabandContext so we can create other objects
    # There should be no Orchestrations tasks.
    # However, let's disable any orchestrations side effects
    config_for_tracking = {
        "run": {
            "skip_completed": False,
            "skip_completed_on_run": False,
            "validate_task_inputs": False,
            "validate_task_outputs": False,
        },  # we don't want to "check" as script is task_version="now"
        "task": {"task_in_memory_outputs": True},  # do not save any outputs
        "core": {"tracker_raise_on_error": False},  # do not fail on tracker errors
    }
    if airflow_context:
        import pytz

        task_target_date = pendulum.parse(
            airflow_context.execution_date, tz=pytz.UTC
        ).date()
        use_dbnd_log = override_airflow_log_system_for_tracking()
        if use_dbnd_log is not None:
            config_for_tracking["log"] = {"disabled": not use_dbnd_log}

        config_for_tracking["task"]["task_target_date"] = task_target_date

    return config.set_values(
        config_values=config_for_tracking,
        priority=ConfigValuePriority.OVERRIDE,
        source="dbnd_tracking_config",
    )


def _set_process_exit_handler(handler):
    atexit.register(handler)

    # https://docs.python.org/3/library/atexit.html
    # The functions registered via this module are not called when the program
    # is killed by a signal not handled by Python, when a Python fatal internal
    # error is detected, or when os._exit() is called.
    #                       ^^^^^^^^^^^^^^^^^^^^^^^^^
    # and os._exit is the one used by airflow (and maybe other libraries)
    # so we'd like to monkey-patch os._exit to stop dbnd inplace run manager
    original_os_exit = os._exit

    def _dbnd_os_exit(*args, **kwargs):
        try:
            handler()
        finally:
            original_os_exit(*args, **kwargs)

    os._exit = _dbnd_os_exit


def _build_inline_root_task(root_task_name):
    # create "root task" with default name as current process executable file name

    task_definition = TaskDefinition(
        task_passport=TaskPassport.from_module(
            TrackingTask.__module__
        ),  # we need to fix that
        source_code=TaskSourceCode.from_callstack(),
    )

    root_task = TrackingTask(
        task_name=root_task_name,
        task_definition=task_definition,
        task_params=Parameters(source="inline_root_task", param_values=[]),
    )

    root_task.ctrl.task_repr.task_command_line = list2cmdline(sys.argv)
    root_task.ctrl.task_repr.task_functional_call = "bash_cmd(args=%s)" % repr(sys.argv)

    return root_task


def _set_dbnd_config_from_airflow_connections():
    """Set Databand config from Extra section in Airflow dbnd_config connection."""
    try:
        from dbnd_airflow.tracking.dbnd_airflow_conf import (
            set_dbnd_config_from_airflow_connections,
        )

        set_dbnd_config_from_airflow_connections()

    except ImportError:
        logger.info(
            "dbnd_airflow is not installed. Config will not load from Airflow Connections"
        )


def has_level_handler(loggr: logging.Logger):
    """Check if there is a handler in the logging chain that will handle the
    given logger's :meth:`effective level <~logging.Logger.getEffectiveLevel>`.

    Arguments:
        loggr : the logger to inspect
    """
    level = loggr.getEffectiveLevel()
    current = loggr

    while current:
        if any(handler.level <= level for handler in current.handlers):
            return True

        if not current.propagate:
            break

        current = current.parent

    return False


def _configure_tracking_logging(conf):
    dbnd_logger = logging.getLogger("dbnd")
    logging_level = conf["log"].get("level", "WARNING")
    dbnd_logger.setLevel(logging_level)
    if not has_level_handler(dbnd_logger):
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(logging.Formatter(FORMAT_SIMPLE))
        dbnd_logger.addHandler(console_handler)


class _DbndScriptTrackingManager(object):
    def __init__(self):
        self._context_managers = []
        self._atexit_registered = False

        self._active = False

        self._run = None
        self._task_run = None

    def _enter_cm(self, cm):
        # type: (typing.ContextManager[T]) -> T
        # else contextManagers are getting closed sometimes :(
        val = cm.__enter__()
        self._context_managers.append(cm)
        return val

    def _close_all_context_managers(self):
        while self._context_managers:
            cm = self._context_managers.pop()
            cm.__exit__(None, None, None)

    def update_run_from_airflow_context(self, airflow_context):
        if not airflow_context or not airflow_context.context:
            return

        dag = airflow_context.context.get("dag", None)
        if not dag:
            return

        dag_tags = getattr(dag, "tags", [])
        project_name = get_project_name_from_airflow_tags(dag_tags)
        airflow_user = airflow_context.context["dag"].owner

        if project_name:
            self._run.project_name = project_name

        if airflow_user:
            self._run.context.task_run_env.user = airflow_user

        if airflow_context.is_subdag:
            root_run_uid = get_job_run_uid(
                airflow_instance_uid=airflow_context.airflow_instance_uid,
                dag_id=airflow_context.root_dag_id,
                execution_date=airflow_context.execution_date,
            )
            self._run.root_run_info = RootRunInfo(
                root_run_uid=root_run_uid,
                root_task_run_uid=get_task_run_uid(
                    run_uid=root_run_uid,
                    dag_id=airflow_context.root_dag_id,
                    task_id=airflow_context.dag_id.split(".")[-1],
                ),
            )

    def start(self, root_task_name=None, project_name=None, airflow_context=None):
        if self._run or self._active or try_get_databand_run():
            return

        # we probably should use only airlfow context via parameter.
        # also, there are mocks that cover only get_dbnd_project_config().airflow_context
        airflow_context = airflow_context or get_dbnd_project_config().airflow_context()
        if airflow_context:
            _set_dbnd_config_from_airflow_connections()

        _set_tracking_config_overide(airflow_context=airflow_context)
        dc = self._enter_cm(
            new_dbnd_context(name="inplace_tracking")
        )  # type: DatabandContext

        if not root_task_name:
            # extract the name of the script we are running (in Airflow scenario it will be just "airflow")
            root_task_name = sys.argv[0].split(os.path.sep)[-1]

        if airflow_context:
            root_task, job_name, source, run_uid = build_run_time_airflow_task(
                airflow_context, root_task_name
            )
            try_number = airflow_context.try_number
        else:
            root_task = _build_inline_root_task(root_task_name)
            job_name = root_task_name
            source = UpdateSource.generic_tracking
            run_uid = None
            try_number = 1

        tracking_source = (
            None  # TODO_CORE build tracking_source -> typeof TrackingSourceSchema
        )
        self._run = run = self._enter_cm(
            new_databand_run(
                context=dc,
                job_name=job_name,
                run_uid=run_uid,
                existing_run=run_uid is not None,
                source=source,
                af_context=airflow_context,
                tracking_source=tracking_source,
                project_name=project_name,
            )
        )  # type: DatabandRun

        self._run.root_task = root_task

        self.update_run_from_airflow_context(airflow_context)

        if not self._atexit_registered:
            _set_process_exit_handler(self.stop)
            self._atexit_registered = True

        sys.excepthook = self.stop_on_exception
        self._active = True

        # now we send data to DB
        root_task_run = run._build_and_add_task_run(
            root_task, task_af_id=root_task.task_name, try_number=try_number
        )

        root_task_run.is_root = True

        run.tracker.init_run()
        run.root_task_run.set_task_run_state(TaskRunState.RUNNING)

        should_capture_log = TrackingConfig.from_databand_context().capture_tracking_log
        self._enter_cm(
            run.root_task_run.runner.task_run_execution_context(
                capture_log=should_capture_log, handle_sigterm=False
            )
        )
        self._task_run = run.root_task_run

        return self._task_run

    def stop(self, finalize_run=True):
        if not self._active:
            return
        self._active = False
        try:
            # Required for scripts tracking which do not set the state to SUCCESS
            if finalize_run:
                databand_run = self._run
                root_tr = self._task_run
                root_tr.finished_time = utcnow()

                if root_tr.task_run_state not in TaskRunState.finished_states():
                    for tr in databand_run.task_runs:
                        if tr.task_run_state == TaskRunState.FAILED:
                            root_tr.set_task_run_state(TaskRunState.UPSTREAM_FAILED)
                            break
                    else:
                        # We can reach here in case of raising exception tracking stand alone python script
                        if sys.exc_info()[1]:
                            error = TaskRunError.build_from_ex(None, root_tr)
                            root_tr.set_task_run_state(TaskRunState.FAILED, error=error)
                        else:
                            root_tr.set_task_run_state(TaskRunState.SUCCESS)

                if root_tr.task_run_state == TaskRunState.SUCCESS:
                    databand_run.set_run_state(RunState.SUCCESS)
                else:
                    databand_run.set_run_state(RunState.FAILED)
                if root_tr.task_run_state == TaskRunState.DEFERRED:
                    return

            self._close_all_context_managers()

        except Exception:
            _handle_tracking_error("dbnd-tracking-shutdown")

    def stop_on_exception(self, type, value, traceback):
        if self._active:
            try:
                error = TaskRunError.build_from_ex(
                    ex=value, task_run=self._task_run, exc_info=(type, value, traceback)
                )
                self._task_run.set_task_run_state(TaskRunState.FAILED, error=error)
            except:
                _handle_tracking_error("dbnd-set-script-error")

        self.stop()
        sys.__excepthook__(type, value, traceback)


# API functions


def try_get_inplace_tracking_task_run():
    # type: ()->Optional[TaskRun]
    if get_dbnd_project_config().is_tracking_mode():
        return dbnd_tracking_start()


# there can be only one tracking manager
_dbnd_script_manager = None  # type: Optional[_DbndScriptTrackingManager]


def tracking_start_base(job_name, project_name=None, airflow_context=None):
    """
    Starts handler for tracking the current running script.
    Would not start a new one if script manager if already exists
    """
    dbnd_project_config = get_dbnd_project_config()
    if dbnd_project_config.disabled:
        # we are not tracking if dbnd is disabled
        return None

    global _dbnd_script_manager
    if not _dbnd_script_manager:
        # setting the context to tracking to prevent conflicts from dbnd orchestration
        dbnd_project_config._dbnd_tracking = True

        dsm = _DbndScriptTrackingManager()
        try:
            # we use job name for both job_name and root_task_name of the run
            dsm.start(job_name, project_name, airflow_context)
            if dsm._active:
                _dbnd_script_manager = dsm

        except Exception:
            _handle_tracking_error("dbnd-tracking-start")

            # disabling the project so we don't start any new handler in this execution
            dbnd_project_config.disabled = True
            return None

    if _dbnd_script_manager and _dbnd_script_manager._active:
        # this is the root task run of the tracking, its representing the script context.
        return _dbnd_script_manager._task_run


def dbnd_airflow_tracking_start(airflow_context):
    """This function is meant to be used only from inside Airflow for Airflow tracking only."""
    job_name = try_get_script_name()
    return tracking_start_base(job_name=job_name, airflow_context=airflow_context)


def dbnd_tracking_start(job_name=None, run_name=None, project_name=None, conf=None):
    """
    This function is used for tracking Python scripts only and should be added at the beginning of the script.

    When the script execution ends, dbnd_tracking_stop will be called automatically, there is no need to add it manually.

    Args:
        job_name: Name of the pipeline
        run_name: Name of the run
        project_name: Name of the project
        conf: Configuration dict with values for Databand configurations
    """
    if not conf:
        conf = {}

    if run_name:
        conf.setdefault("run", {}).setdefault("name", run_name)

    # do not apply our logger to a python script by default
    conf.setdefault("log", {}).setdefault("disabled", True)

    # send logs to webserver by default
    conf.setdefault("tracking", {}).setdefault("capture_tracking_log", True)

    if conf["log"]["disabled"]:
        _configure_tracking_logging(conf)

    # We use print here and not log because the dbnd logger might be set to Warning (by default), and we want to
    # inform the user that we started, without alerting him with a Warning or Error message.
    # This should be a logger info message when tracking and orchestration split.
    print("Databand Tracking Started {version}".format(version=dbnd.__version__))

    if conf:
        config.set_values(
            config_values=conf,
            priority=ConfigValuePriority.OVERRIDE,
            source="dbnd_tracking_start",
        )

    if job_name is None:
        job_name = try_get_script_name()

    return tracking_start_base(job_name=job_name, project_name=project_name)


def dbnd_tracking_stop(finalize_run=True):
    """
    Stops and clears the script tracking if exists.

    Args:
        finalize_run: Should complete the run by setting it's state to a complete one (success or failed).
    """
    global _dbnd_script_manager
    if _dbnd_script_manager:
        _dbnd_script_manager.stop(finalize_run)
        _dbnd_script_manager = None


def is_dbnd_tracking_active() -> bool:
    global _dbnd_script_manager
    return bool(_dbnd_script_manager)


@seven.contextlib.contextmanager
def dbnd_tracking(job_name=None, run_name=None, project_name=None, conf=None):
    # type: (...) -> TaskRun
    """
    This function is used for tracking Python scripts only and should be used with a with statement.

    Args:
        job_name: Name of the pipeline
        run_name: Name of the run
        project_name: Name of the project
        conf: Configuration dict with values for Databand configurations
    """
    try:
        tr = dbnd_tracking_start(
            job_name=job_name, run_name=run_name, project_name=project_name, conf=conf
        )
        yield tr
    finally:
        dbnd_tracking_stop()



# © Copyright Databand.ai, an IBM Company 2022

from typing import TypeVar, Union

import pytest

from dbnd import Task, dbnd_run_cmd
from dbnd._core.run.databand_run import DatabandRun
from dbnd._core.utils.platform import windows_compatible_mode


T = TypeVar("T")


def run_locally__raises(expected_exception, args, conf=None):
    import pytest

    with pytest.raises(expected_exception):
        dbnd_run_cmd(args)


def assert_run_task(task):  # type: (Union[T, Task]) -> Union[T, Task]
    task.dbnd_run()
    assert task._complete()
    return task


def assert_run(task):  # type: (Union[T, Task]) -> DatabandRun
    run = task.dbnd_run()
    assert task._complete()
    return run


skip_on_windows = pytest.mark.skipif(
    windows_compatible_mode, reason="not supported on Windows OS"
)



# © Copyright Databand.ai, an IBM Company 2022

import inspect
import logging

from typing import Callable, Optional, Type

import attr

from dbnd._core.errors.errors_utils import UserCodeDetector
from dbnd._core.task.base_task import _BaseTask


logger = logging.getLogger(__name__)


def _hide_or_show_by_config(source):
    from dbnd._core.settings import TrackingConfig

    tracking_config = TrackingConfig.from_databand_context()
    if tracking_config.track_source_code:
        return source
    else:
        return None


@attr.s
class TaskSourceCode(object):
    _task_source_code = attr.ib(default=None)  # type: Optional[str]
    _task_module_code = attr.ib(default=None)  # type: Optional[str]
    _task_source_file = attr.ib(default=None)  # type: Optional[str]

    @property
    def task_source_code(self):
        return _hide_or_show_by_config(source=self._task_source_code)

    @task_source_code.setter
    def task_source_code(self, value):
        self._task_source_code = value

    @property
    def task_module_code(self):
        return _hide_or_show_by_config(source=self._task_module_code)

    @task_module_code.setter
    def task_module_code(self, value):
        self._task_module_code = value

    @property
    def task_source_file(self):
        return _hide_or_show_by_config(source=self._task_source_file)

    @task_source_file.setter
    def task_source_file(self, value):
        self._task_source_file = value

    @classmethod
    def from_callable(cls, callable):
        # type: (Callable) -> TaskSourceCode
        task_source_code = _get_source_code(callable)
        task_module_code = _get_module_source_code(callable)
        task_source_file = _get_source_file(callable)

        return TaskSourceCode(
            task_source_code=task_source_code,
            task_module_code=task_module_code,
            task_source_file=task_source_file,
        )

    @classmethod
    def from_class(cls, class_obj):
        # type: (type) -> TaskSourceCode
        task_source_code = _get_source_code(class_obj)
        task_module_code = _get_module_source_code(class_obj)
        task_source_file = _get_source_file(class_obj.__class__)

        return TaskSourceCode(
            task_source_code=task_source_code,
            task_module_code=task_module_code,
            task_source_file=task_source_file,
        )

    @classmethod
    def from_task_class(cls, task_class):
        # type: (Type[_BaseTask]) -> TaskSourceCode
        # TODO: FIX TASK_DECORATOR
        if task_class._conf__track_source_code:
            if hasattr(task_class, "task_decorator") and task_class.task_decorator:
                # this means we are inside a task_class of decorated function
                return cls.from_callable(task_class.task_decorator.class_or_func)
            else:
                return cls.from_class(task_class)
        else:
            return NO_SOURCE_CODE

    @classmethod
    def from_callstack(cls):
        try:
            user_frame = UserCodeDetector.build_code_detector().find_user_side_frame(
                user_side_only=True
            )
            if user_frame:
                module_code = open(user_frame.filename).read()
                return TaskSourceCode(
                    task_module_code=module_code, task_source_file=user_frame.filename
                )
        except Exception as ex:
            logger.debug("Failed to find source code: %s", str(ex))
        return NO_SOURCE_CODE


NO_SOURCE_CODE = TaskSourceCode()


def _get_source_code(item):
    try:
        import inspect

        return inspect.getsource(item)
    except (TypeError, OSError):
        logger.debug("Failed to task source for %s", item)
    except Exception:
        logger.debug("Error while getting task source")
    return "Error while getting source code"


def _get_module_source_code(item):
    try:
        return inspect.getsource(inspect.getmodule(item))
    except TypeError:
        logger.debug("Failed to module source for %s", item)
    except Exception:
        logger.exception("Error while getting module source")
    return "Error while getting source code"


def _get_source_file(item):
    try:
        return inspect.getfile(item).replace(".pyc", ".py")
    except Exception:
        logger.warning("Failed find a path of source code for task {}".format(item))
    return None



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.errors import DatabandError, DatabandRuntimeError
from dbnd._core.errors.friendly_error.helpers import _run_name
from dbnd._core.utils.console_utils import bold, cyan, red, underline


def dataflow_pipeline_not_set(task):
    return DatabandRuntimeError(
        "dataflow_pipeline at {task} is None. Can't wait on dataflow job completion.".format(
            task=_run_name(task)
        ),
        help_msg="Please set task.pipeline first at '{task}' or change task.dataflow_wait_until_finish flag".format(
            task=_run_name(task)
        ),
    )


def logger_format_for_databand_error(error):
    # type: (DatabandError) -> str

    title_error = red(underline(bold("Error: ")))
    title_exception = red(underline(bold("Nested exception: ")))
    title_helper = cyan(underline(bold("Help: ")))

    full_err_message = "\n\n {title_error}{msg_error} \n {title_exception}{msg_exception} \n {title_helper}\n{msg_helper}\n\n".format(
        title_error=title_error,
        msg_error=error,
        title_exception=title_exception,
        msg_exception=getattr(error, "nested_exceptions", ""),
        title_helper=title_helper,
        msg_helper=getattr(error, "help_msg", ""),
    )
    return full_err_message



# © Copyright Databand.ai, an IBM Company 2022

import os

from os.path import normpath
from typing import Optional

import attr
import six

from targets.utils.path import trailing_slash


class FileCompressions(object):
    none = "none"
    gzip = "gz"
    bzip = "bz2"
    zip = "zip"
    xz = "xz"
    snappy = "snappy"


class FileFormat(object):
    csv = "csv"
    txt = "txt"
    hdf5 = "h5"
    json = "json"
    table = "table"
    feather = "feather"
    numpy = "npy"
    jpeg = "jpeg"
    png = "png"
    pdf = "pdf"
    excel = "xlsx"
    parquet = "parquet"
    html = "html"
    pickle = "pickle"
    yaml = "yaml"
    tsv = "tsv"
    tfmodel = "tfmodel"
    tfhistory = "tfhistory"
    zip = "zip"


BINARY_FORMATS = [
    FileFormat.hdf5,
    FileFormat.feather,
    FileFormat.numpy,
    FileFormat.excel,
    FileFormat.parquet,
    FileFormat.pickle,
    FileFormat.tfmodel,
]


@attr.s(frozen=True, repr=True)
class _FileExtensionMap(object):
    _ext_to_name = attr.ib(factory=dict)
    _name_to_ext = attr.ib(factory=dict)

    def register_extension(self, name, extensions=None):
        extensions = extensions or name
        if isinstance(extensions, six.string_types):
            extensions = [extensions]
        for e in extensions:
            self._ext_to_name[e] = name
        self._name_to_ext[name] = extensions

    def get_default_extension(self, name):
        if not name:
            return ""

        if name in self._name_to_ext:
            ext = self._name_to_ext[name][0]
            return ".%s" % ext if ext else ""

        # default fallback
        return ".%s" % name

    def known_ext(self, name):
        return name in self._ext_to_name

    def extension_to_name(self, extention):
        return self._ext_to_name.get(extention)


KNOWN_FORMATS = _FileExtensionMap()
KNOWN_COMPRESSIONS = _FileExtensionMap()


def register_file_extension(name, extra_extensions=None):
    KNOWN_FORMATS.register_extension(name, extra_extensions)
    return name


def register_compression(name, ext=None):
    KNOWN_COMPRESSIONS.register_extension(name, ext)


register_compression(FileCompressions.gzip, "gz")
register_compression(FileCompressions.bzip, "bz2")
register_compression(FileCompressions.zip)
register_compression(FileCompressions.xz)
register_compression(FileCompressions.snappy)

register_file_extension(FileFormat.csv)
register_file_extension(FileFormat.txt, ["txt", "py", "log"])
register_file_extension(FileFormat.table)
register_file_extension(FileFormat.parquet)
register_file_extension(FileFormat.feather)
register_file_extension(FileFormat.excel)
register_file_extension(FileFormat.pickle)
register_file_extension(FileFormat.html)
register_file_extension(FileFormat.png)
register_file_extension(FileFormat.pdf)
register_file_extension(FileFormat.jpeg, ["jpg", "jpeg"])
register_file_extension(FileFormat.hdf5, ["h5", "hdf5"])
register_file_extension(FileFormat.json, ["json", "hjson"])
register_file_extension(FileFormat.yaml, ["yaml", "yml"])
register_file_extension(FileFormat.numpy, ["npy", "numpy"])
register_file_extension(FileFormat.tsv)
register_file_extension(FileFormat.tfmodel)
register_file_extension(FileFormat.tfhistory)
register_file_extension(FileFormat.zip)


@attr.s(frozen=True, repr=False)
class TargetConfig(object):
    format = attr.ib(default=None)  # type: Optional[str]
    compression = attr.ib(default=None)  # type: Optional[str]
    folder = attr.ib(default=False)  # type: bool
    meta_files = attr.ib(default=tuple())

    flag = attr.ib(default=True)
    is_binary = attr.ib(default=False)
    target_factory = attr.ib(default=None)
    require_local_access = attr.ib(default=False)
    overwrite_target = attr.ib(default=False)

    def with_compression(self, compression):
        return attr.evolve(self, compression=compression)

    def with_format(self, format):
        # type: (TargetConfig, str) -> TargetConfig
        return attr.evolve(self, format=format, is_binary=format in BINARY_FORMATS)

    def with_meta_files(self, list_of_meta_files):
        return attr.evolve(self, meta_files=list_of_meta_files, folder=True)

    def with_flag(self, flag):
        # type: (TargetConfig, bool) -> TargetConfig
        return attr.evolve(self, flag=flag)

    def without_flag(self):
        return attr.evolve(self, flag=None)

    def with_target_factory(self, target_factory):
        return attr.evolve(self, target_factory=target_factory)

    def as_folder(self):
        return attr.evolve(self, folder=True)

    def as_file(self):
        return attr.evolve(self, folder=False, flag=None, target_factory=None)

    def with_require_local_access(self):
        return attr.evolve(self, require_local_access=True)

    def with_overwrite(self):
        return attr.evolve(self, overwrite_target=True)

    @property
    def gzip(self):
        return self.with_compression(FileCompressions.gzip)

    @property
    def gz(self):
        return self.gzip

    @property
    def bzip(self):
        return self.with_compression(FileCompressions.bzip)

    @property
    def snappy(self):
        return self.with_compression(FileCompressions.snappy)

    @property
    def zip(self):
        return self.with_format(FileFormat.zip)

    @property
    def xz(self):
        return self.with_compression(FileCompressions.xz)

    @property
    def csv(self):
        return self.with_format(FileFormat.csv)

    @property
    def tsv(self):
        return self.with_format(FileFormat.tsv)

    @property
    def txt(self):
        return self.with_format(FileFormat.txt)

    @property
    def hdf5(self):
        return self.with_format(FileFormat.hdf5)

    @property
    def json(self):
        return self.with_format(FileFormat.json)

    @property
    def feather(self):
        return self.with_format(FileFormat.feather)

    @property
    def numpy(self):
        return self.with_format(FileFormat.numpy)

    @property
    def excel(self):
        return self.with_format(FileFormat.excel)

    @property
    def parquet(self):
        return self.with_format(FileFormat.parquet)

    @property
    def pickle(self):
        return self.with_format(FileFormat.pickle)

    @property
    def tfmodel(self):
        return self.with_format(FileFormat.tfmodel)

    @property
    def tfhistory(self):
        return self.with_format(FileFormat.tfhistory)

    @property
    def overwrite(self):
        return self.with_overwrite()

    def get_ext(self):  # type: (TargetConfig) -> str
        ext = ""
        if self.format:
            ext += KNOWN_FORMATS.get_default_extension(self.format)

        if self.format not in [
            file.parquet,
            file.hdf5,
        ]:  # we don't add compression to this formats
            ext += KNOWN_FORMATS.get_default_extension(self.compression)
        return ext

    def __repr__(self):
        ext = self.get_ext()
        if self.folder:
            ext += "/"
        return ext


file = TargetConfig()
folder = TargetConfig().as_folder()


def parse_target_config(config_str):
    if not config_str.startswith("."):
        config_str = ".%s" % config_str
    return extract_target_config_from_path(config_str)


def extract_target_config_from_path(path, config=None):
    # type: (str, TargetConfig)-> TargetConfig
    config = config or TargetConfig()
    if path is None:
        return config
    if path.endswith("[noflag]"):
        config = config.with_flag(None)
        path = path[:-8]

    if trailing_slash(path):
        # we move to folder mode regardles config
        config = config.as_folder()
        path = os.path.dirname(path)

    path = normpath(path)
    if config.format or config.compression:
        # we don't set "forced" settings
        return config
    path = os.path.basename(path)

    # Infer format and compression from the filename/URL extension
    parts = path.split(".")
    if len(parts) > 2:
        parts = parts[-2:]
    for p in parts:
        # priority to last elements
        # for example  parquet.csv is csv
        format = KNOWN_FORMATS.extension_to_name(p)
        if format:
            config = config.with_format(format)
        else:
            compression = KNOWN_COMPRESSIONS.extension_to_name(p)
            if compression:
                config = config.with_compression(compression)
    return config



# © Copyright Databand.ai, an IBM Company 2022

from collections import namedtuple


CONF_TASK_SECTION = "task"
CONF_CONFIG_SECTION = "config"

ConfigPath = namedtuple("ConfigPath", ["section", "key"])



# © Copyright Databand.ai, an IBM Company 2022

import sys
import traceback


def format_exception_as_str(exc_info=None):
    if not exc_info:
        exc_info = sys.exc_info()

    exc_type, exc_value, exc_traceback = exc_info
    lines = traceback.format_exception(
        type(exc_value), exc_value, exc_traceback, limit=None
    )
    return "".join(lines)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.current import get_databand_context
from dbnd.api.shared_schemas.airflow_monitor import AirflowServerInfoSchema


def list_synced_airflow_instances():
    endpoint = "airflow_monitor"
    schema = AirflowServerInfoSchema(strict=False)
    client = get_databand_context().databand_api_client
    response = client.api_request(endpoint=endpoint, data="", method="GET")
    return schema.load(data=response["data"], many=True).data


def generate_access_token(name, lifespan):
    client = get_databand_context().databand_api_client
    resp = client.api_request(
        "/api/v1/auth/personal_access_token",
        {"label": name, "lifespan": lifespan},
        method="POST",
    )
    return resp


def create_airflow_instance(
    url,
    external_url,
    fetcher,
    env,
    dag_ids,
    last_seen_dag_run_id,
    last_seen_log_id,
    name,
    generate_token,
    system_alert_definitions,
    monitor_config,
):
    client = get_databand_context().databand_api_client
    endpoint = "airflow_monitor/add"
    request_data = {
        "base_url": url,
        "external_url": external_url,
        "fetcher": fetcher,
        "env": env,
        "monitor_config": monitor_config,
        "dag_ids": dag_ids,
        "last_seen_dag_run_id": last_seen_dag_run_id,
        "last_seen_log_id": last_seen_log_id,
        "name": name,
        "system_alert_definitions": system_alert_definitions,
    }

    if monitor_config:
        request_data["monitor_config"] = monitor_config

    resp = client.api_request(endpoint, request_data, method="POST")
    config_json = resp["server_info_dict"]
    config_json["core"][
        "databand_url"
    ] = get_databand_context().settings.core.databand_url

    if generate_token:
        token_resp = generate_access_token(name, generate_token)

        config_json["core"]["databand_access_token"] = token_resp["token"]

    return config_json


def edit_airflow_instance(
    tracking_source_uid,
    url,
    external_url,
    fetcher,
    env,
    dag_ids,
    last_seen_dag_run_id,
    last_seen_log_id,
    name,
    system_alert_definitions,
    monitor_config,
):
    client = get_databand_context().databand_api_client
    endpoint = "airflow_monitor/edit"
    request_data = {
        "tracking_source_uid": tracking_source_uid,
        "base_url": url,
        "external_url": external_url,
        "fetcher": fetcher,
        "env": env,
        "monitor_config": monitor_config,
        "dag_ids": dag_ids,
        "last_seen_dag_run_id": last_seen_dag_run_id,
        "last_seen_log_id": last_seen_log_id,
        "name": name,
        "system_alert_definitions": system_alert_definitions,
    }

    client.api_request(endpoint, request_data, method="POST")


def archive_airflow_instance(url):
    client = get_databand_context().databand_api_client
    endpoint = "airflow_monitor/archive"
    client.api_request(endpoint, url, method="POST")


def unarchive_airflow_instance(url):
    client = get_databand_context().databand_api_client
    endpoint = "airflow_monitor/unarchive"
    client.api_request(endpoint, url, method="POST")



# © Copyright Databand.ai, an IBM Company 2022

from collections import defaultdict
from typing import Any, Dict

import attr
import six

from targets import InMemoryTarget, Target
from targets.values import ValueType


@attr.s
class ValueOrigin(object):
    obj_id = attr.ib()  # type: int  # object id()
    origin_target = attr.ib()  # type: Target
    value_type = attr.ib()  # type: ValueType


class TargetIdentitySourceMap(object):
    """
    Used to store map from target's value object (id()) to target it came from (+ parameter).
    Usually will be stored on DatabandRun level.
    When using dynamic (inline) tasks - used to know origin target some value came from.
    """

    def __init__(self):
        self.id_map = defaultdict(list)

    def add(self, origin_target, runtime_value, value_type):
        # type: (Target, Any, ValueType) -> None
        # https://stackoverflow.com/questions/306313/is-operator-behaves-unexpectedly-with-integers
        # or should we just exclude all integers? what about other types? like:
        # http://guilload.com/python-string-interning/
        if (
            runtime_value is None
            or _is_cached_integer(runtime_value)
            or not isinstance(origin_target, Target)
        ):
            return

        obj_id = id(runtime_value)
        if isinstance(origin_target, InMemoryTarget):
            value_type = origin_target.value_type

        self.id_map[obj_id].append(
            ValueOrigin(
                obj_id=obj_id, origin_target=origin_target, value_type=value_type
            )
        )

    def get(self, v):
        # type: (Any) -> ValueOrigin
        return self.id_map.get(id(v))

    def get_for_map(self, d):
        # type: (Dict[str, Any]) -> Dict[str, ValueOrigin]
        return {
            key: self.id_map[id(value)][0]
            for key, value in six.iteritems(d)
            if id(value) in self.id_map
        }


def _is_cached_integer(value):
    return isinstance(value, six.integer_types) and (-5 <= value <= 256)



# © Copyright Databand.ai, an IBM Company 2022

import typing

from collections import Counter

from dbnd._core.constants import DescribeFormat, SystemTaskName, TaskRunState
from dbnd._core.current import is_verbose
from dbnd._core.run.run_ctrl import RunCtrl
from dbnd._core.tracking.schemas.tracking_info_objects import TaskRunEnvInfo
from dbnd._core.utils.basics.text_banner import TextBanner


if typing.TYPE_CHECKING:
    from dbnd._core.run.databand_run import DatabandRun


class RunBanner(RunCtrl):
    @property
    def tracker(self):
        return self.run.tracker

    def _add_tasks_info(self, b):
        run = self.run
        reused = sum(
            tr.is_reused and not tr.is_skipped_as_not_required for tr in run.task_runs
        )
        optimizations = []
        if reused:
            optimizations.append(
                "There are {completed} reused tasks.".format(completed=reused)
            )
        task_skipped_as_not_required = sum(
            tr.is_reused and tr.is_skipped_as_not_required for tr in run.task_runs
        )
        if task_skipped_as_not_required:
            optimizations.append(
                " {skipped} tasks are not required by any uncompleted task "
                "that is essential for your root task.".format(
                    skipped=task_skipped_as_not_required
                )
            )
        show_more = is_verbose()
        task_runs = run.task_runs
        if not show_more:  # show only non system
            task_runs = run.get_task_runs(without_system=True)
            if not task_runs:  # we have only system
                task_runs = run.task_runs

        states = Counter(tr.task_run_state for tr in task_runs if tr.task_run_state)

        b.column("TOTAL TASKS", len(task_runs))

        tasks = [(k.value, v) for k, v in states.items()]
        b.column_properties("STATES", tasks)
        if optimizations:
            b.column("RUN OPTIMIZATION", " ".join(optimizations))
        return b

    def run_banner(
        self, msg, color="white", show_run_info=False, show_tasks_info=False
    ):
        run = self.run  # type: DatabandRun
        if run.root_run_info.root_run_uid != run.run_uid:
            msg += " -> sub-run"

        b = TextBanner(msg, color)
        ctx = run.context
        task_run_env = ctx.task_run_env  # type: TaskRunEnvInfo

        if run.context.tracking_store.has_tracking_store("api", channel_name="web"):
            b.column("TRACKER URL", run.run_url, skip_if_empty=True)

        b.column("TRACKERS", run.context.tracking_store.trackers_names)

        if run.root_run_info.root_run_uid != run.run_uid:
            b.column(
                "ROOT TRACKER URL", run.root_run_info.root_run_url, skip_if_empty=True
            )
            b.column("ROOT UID URL", run.root_run_info.root_run_uid, skip_if_empty=True)

        if run.scheduled_run_info:
            b.column_properties(
                "SCHEDULED",
                [
                    ("scheduled_job", run.scheduled_run_info.scheduled_job_uid),
                    ("scheduled_date", run.scheduled_run_info.scheduled_date),
                    ("dag_run_id", run.scheduled_run_info.scheduled_job_dag_run_id),
                ],
            )

        if show_run_info:
            b.new_line()
            run_params = [
                ("user", task_run_env.user),
                ("run_uid", "%s" % run.run_uid),
                ("env", run.env.name),
                ("project", run.project_name) if run.project_name else None,
                ("user_code_version", task_run_env.user_code_version),
            ]
            run_params = list(filter(None, run_params))
            b.column("RUN", b.f_simple_dict(run_params))
            b.column("CMD", task_run_env.cmd_line)

            if task_run_env.user_data and task_run_env.user_data != "None":
                b.column("USER DATA", task_run_env.user_data, skip_if_empty=True)
            b.new_line()

        if run.is_orchestration:
            run_executor = run.run_executor
            driver_task_run = run.driver_task_run
            if show_run_info:
                if driver_task_run and driver_task_run.log:
                    b.column(
                        "LOG",
                        b.f_simple_dict(
                            [
                                ("local", driver_task_run.log.local_log_file),
                                ("remote", driver_task_run.log.remote_log_file),
                            ],
                            skip_if_empty=True,
                        ),
                    )

                b.column(
                    "EXECUTE",
                    b.f_simple_dict(
                        [
                            ("TASK_EXECUTOR", run_executor.task_executor_type),
                            ("PARALLEL", run_executor.parallel),
                            ("SUBMIT_DRIVER", run_executor.submit_driver),
                            ("SUBMIT_TASKS", run_executor.submit_tasks),
                        ],
                        skip_if_empty=True,
                    ),
                    skip_if_empty=True,
                )
            if run_executor.run_executor_type == SystemTaskName.driver:
                if run.root_task_run:
                    b.column("TASK_BAND", run.root_task_run.task.task_band)

                if show_tasks_info:
                    self._add_tasks_info(b)
                failed_task_runs = [
                    task_run
                    for task_run in run.get_task_runs()
                    if task_run.task_run_state == TaskRunState.FAILED
                ]
                if failed_task_runs:
                    f_msg = "\n".join(tr.task.task_id for tr in failed_task_runs)
                    b.column("FAILED", f_msg)
        b.new_line()

        return b.getvalue()

    def run_banner_for_submitted(self):
        msg = (
            "Your run has been submitted! Please check submitted job in UI\n"
            " * Use '--interactive' to have blocking run (wait for driver completion)\n"
            " * Use '--local-driver' (env.submit_driver=False) to run your driver locally\n"
        )
        return self.run_banner(
            msg, color="green", show_run_info=False, show_tasks_info=False
        )


def print_tasks_tree(root_task, task_runs, describe_format=DescribeFormat.short):
    from dbnd._core.task_ctrl.task_dag_describe import DescribeDagCtrl

    completed = {tr.task.task_id: tr.is_reused for tr in task_runs}
    run_describe_dag = DescribeDagCtrl(
        root_task, describe_format, complete_status=completed
    )
    run_describe_dag.tree_view(describe_format=describe_format)



# © Copyright Databand.ai, an IBM Company 2022

import json
import logging

from decimal import Decimal

import attr
import six

from dbnd._core.constants import _DbndDataClass
from dbnd._core.utils import json_utils
from dbnd._core.utils.string_utils import safe_short_string


logger = logging.getLogger(__name__)


def safe_value_to_str(value, max_len):
    if value is None or isinstance(value, six.string_types):
        return safe_short_string(str(value), max_len)

    try:
        # local import to prevent imports recursion
        from targets.values import get_value_type_of_obj

        value_type = get_value_type_of_obj(value)

        if value_type is not None:
            return value_type.to_preview(value, max_len)
    except Exception:
        logger.warning("failed dumping metric, falling back to str", exc_info=True)
        # This object has to hold only serializable values
    return safe_short_string(str(value), max_len)


class Metric(_DbndDataClass):
    _int_precision = 16
    # Workaround for precission issues with large ints
    MAX_INT = 10 ** (_int_precision - 1)
    MIN_INT = -MAX_INT

    VALUE_MAX_LEN = 5000

    def __init__(
        self,
        key,
        timestamp,
        value_int=None,
        value_float=None,
        value_str=None,
        value_json=None,
        value=None,
        source=None,
    ):
        self.key = key
        self.timestamp = timestamp
        self.source = source
        self.value_int = value_int
        self.value_float = value_float
        self.value_str = value_str
        self.value_json = None
        if value_json is not None:
            try:
                self.value_json = json.loads(json_utils.dumps_safe(value_json))
                self.value_str = None
            except Exception:
                self.value_json = None
                self.value_str = safe_value_to_str(value_json, self.VALUE_MAX_LEN)
        elif (value_int, value_float, value_str) == (None, None, None):
            self.value = value

    @property
    def value(self):
        if self.value_float is not None:
            return self.value_float
        if self.value_int is not None:
            return int(self.value_int)
        if self.value_json is not None:
            return self.value_json
        return self.value_str

    @property
    def serialized_value(self):
        if self.value_json is not None:
            return json.dumps(self.value_json, sort_keys=True)
        else:
            return str(self.value)

    @value.setter
    def value(self, value):
        if isinstance(value, Decimal):
            self.value_float = float(value)
        elif isinstance(value, float):
            self.value_float = value
        elif isinstance(value, int) and self.MIN_INT <= value <= self.MAX_INT:
            # need int() in case it's bool
            self.value_int = int(value)
        else:
            # if passed value has json-serializable representation, we'd prefer
            # to store it as value_json. For this we'll try to convert value to
            # preview (to support max length). If resulting object is valid json
            # then we'll store it. If json is not valid or we have any other
            # error, we'll fallback to just converting it to string
            value_str = safe_value_to_str(value, self.VALUE_MAX_LEN)

            try:
                self.value_json = json.loads(value_str)
            except Exception:
                self.value_str = value_str

    def __repr__(self):
        return "Metric(key={}, source={})".format(self.key, self.source)


@attr.s
class Artifact(_DbndDataClass):
    path = attr.ib()  # type: str



# © Copyright Databand.ai, an IBM Company 2022

import logging
import time
import typing

from typing import Any, Dict, List, Optional, Union

from dbnd._core.constants import DbndDatasetOperationType, DbndTargetOperationType
from dbnd._core.plugin.dbnd_plugins import is_plugin_enabled
from dbnd._core.task_run.task_run_tracker import DatasetOperationReport, TaskRunTracker
from dbnd._core.tracking.log_data_request import LogDataRequest
from dbnd._core.utils import seven
from dbnd._core.utils.one_time_logger import get_one_time_logger
from targets import Target
from targets.value_meta import ValueMetaConf


LOG_DATASET_OP_OP_SOURCE = "python_manual_logging"

if typing.TYPE_CHECKING:
    from datetime import datetime

    import pandas as pd
    import pyspark.sql as spark

    from dbnd_postgres.postgres_values import PostgresTable
    from dbnd_snowflake.snowflake_values import SnowflakeTable

logger = logging.getLogger(__name__)

TRACKER_MISSING_MESSAGE = (
    "Can't report %s because no tracker was found (did you use dbnd_tracking?)"
)


def _get_tracker():
    # type: ()-> Optional[TaskRunTracker]
    """
    Look for a tracker of running task_run or initiate a task_run if nothing is running.
    Will return a None if there is no exist task_run nor couldn't start one.
    """
    from dbnd._core.task_run.current_task_run import try_get_or_create_task_run

    task_run = try_get_or_create_task_run()
    if task_run:
        return task_run.tracker
    return None


def log_data(
    key,  # type: str
    value=None,  # type: Union[pd.DataFrame, spark.DataFrame, PostgresTable, SnowflakeTable]
    path=None,  # type: Optional[str]
    operation_type=DbndTargetOperationType.read,  # type: DbndTargetOperationType
    with_preview=None,  # type: Optional[bool]
    with_size=None,  # type: Optional[bool]
    with_schema=None,  # type: Optional[bool]
    with_stats=None,  # type: Optional[Union[bool, str, List[str], LogDataRequest]]
    with_histograms=None,  # type: Optional[Union[bool, str, List[str], LogDataRequest]]
    raise_on_error=False,  # type: bool
):  # type: (...) -> None
    """
    Log data information to dbnd.

    @param key: Name of the data.
    @param value: Value of the data, currently supporting only dataframes and tables view.
    @param path: Optional target or path representing a target to connect the data to.
    @param operation_type: Type of the operation doing with the target - reading or writing the data?
    @param with_preview: True if should log a preview of the data.
    @param with_size: True if should log the size of the data.
    @param with_schema: True if should log the schema of the data.
    @param with_stats: True if should calculate and log stats of the data.
    @param with_histograms: True if should calculate and log histogram of the data.
    @param raise_on_error: raise if error occur.
    """
    tracker = _get_tracker()
    if not tracker:
        message = TRACKER_MISSING_MESSAGE % ("log_data",)
        get_one_time_logger().log_once(message, "log_data", logging.WARNING)
        return

    meta_conf = ValueMetaConf(
        log_preview=with_preview,
        log_schema=with_schema,
        log_size=with_size,
        log_stats=with_stats,
        log_histograms=with_histograms,
    )

    tracker.log_data(
        key,
        value,
        meta_conf=meta_conf,
        path=path,
        operation_type=operation_type,
        raise_on_error=raise_on_error,
    )


# logging dataframe is the same as logging data
log_dataframe = log_data
log_dataframe.__doc__ = """
    Logs a dataframe to dbnd.

    Args:
        key: Name of the dataframe.
        value: The dataframe itself.
        path: Optional target or path representing a target to connect the dataframe to.
        operation_type: Type of the operation doing with the target - reading or writing the dataframe?
        with_preview: True if should log a preview of the dataframe.
        with_size: True if should log the size of the dataframe.
        with_schema: True if should log the schema of the dataframe.
        with_stats: True if should calculate and log stats of the dataframe.
        with_histograms: True if should calculate and log histogram of the dataframe.
        raise_on_error: raise if error occur.

    Example::

        @task
        def process_customers_data(data) -> pd.DataFrame:
            log_dataframe("customers_data", data)
"""


def log_pg_table(
    table_name,
    connection_string,
    with_preview=None,  # type: Optional[bool]
    with_schema=None,  # type: Optional[bool]
    with_histograms=None,  # type: Union[LogDataRequest, bool, str, List[str]]
):
    """
    Log the data of postgres table to dbnd.

    @param table_name: name of the table to log
    @param connection_string: a connection string used to reach the table.
    @param with_preview: True if should log a preview of the table.
    @param with_schema: True if should log the schema of the table.
    @param with_histograms: True if should calculate and log histogram of the table data.
    """
    try:
        if not is_plugin_enabled("dbnd-postgres", module_import="dbnd_postgres"):
            logger.warning(
                "Can't log postgres table: dbnd-postgres package is not installed\n"
                "Help: pip install dbnd-postgres"
            )
            return

        from dbnd_postgres import postgres_values

        pg_table = postgres_values.PostgresTable(table_name, connection_string)
        log_data(
            table_name,
            pg_table,
            with_preview=with_preview,
            with_schema=with_schema,
            with_histograms=with_histograms,
        )
    except Exception:
        logger.exception("Failed to log_pg_table")


def log_metric(key, value, source="user"):
    # type: (str, Any, Optional[str]) -> None
    """
    Log key-value pair as a metric to dbnd.

    Args:
        key: Name of the metric.
        value: Value of the metric.
        source: The source of the metric, default is user.

    Example::

        def calculate_alpha(alpha):
            alpha *= 1.1
            log_metric("alpha", alpha)
    """
    tracker = _get_tracker()
    if tracker:
        tracker.log_metric(key, value, source=source)
        return

    message = TRACKER_MISSING_MESSAGE % ("log_metric",)
    get_one_time_logger().log_once(message, "log_metric", logging.WARNING)
    logger.info("Log {} Metric '{}'='{}'".format(source.capitalize(), key, value))
    return


def log_metrics(metrics_dict, source="user", timestamp=None):
    # type: (Dict[str, Any], str, datetime) -> None
    """
    Log multiple key-value pairs as metrics to dbnd.

    Args:
        metrics_dict: Key-value pairs of metrics to log.
        source: Optional name of the metrics source, default is user.
        timestamp: Optional timestamp of the metrics.

    Example::

        @task
        def log_lowercase_letters():
            # all lower alphabet chars -> {"a": 97,..., "z": 122}
            log_metrics({chr(i): i for i in range(97, 123)})
    """
    tracker = _get_tracker()
    if tracker:
        tracker.log_metrics(metrics_dict, source=source, timestamp=timestamp)
        return

    message = TRACKER_MISSING_MESSAGE % ("log_metrics",)
    get_one_time_logger().log_once(message, "log_metrics", logging.WARNING)
    logger.info(
        "Log multiple metrics from {source}: {metrics}".format(
            source=source.capitalize(), metrics=metrics_dict
        )
    )


def log_artifact(key, artifact):
    """
    Log a local file or directory as an artifact of the currently active run.

    Args:
        key: The key by which to log the artifact
        artifact: The artifact to log

    Example::

        def prepare_data(data):
            lorem = "Lorem ipsum dolor sit amet, consectetuer adipiscing elit, sed diam nonummy nibh euismod tincidunt"
            data.write(lorem)
            log_artifact("my_tmp_file", str(data))
    """
    tracker = _get_tracker()
    if tracker:
        tracker.log_artifact(key, artifact)
        return

    message = TRACKER_MISSING_MESSAGE % ("log_artifact",)
    get_one_time_logger().log_once(message, "log_artifact", logging.WARNING)
    logger.info("Artifact %s=%s", key, artifact)


@seven.contextlib.contextmanager
def log_duration(metric_key, source="user"):
    """
    Measure time of function or code block, and log to Databand as a metric.

    Can be used as a decorator and in "with" statement as a context manager.

    Example 1::

        @log_duration("f_time_duration")
        def f():
            sleep(1)

    Example 2::

        with log_duration("my_code_duration"):
            sleep(1)
    """
    start_time = time.time()
    try:
        yield
    finally:
        end_time = time.time()
        log_metric(metric_key, end_time - start_time, source)


def _report_operation(operation_report):
    # type: (DatasetOperationReport) -> None
    tracker = _get_tracker()
    if not tracker:
        message = TRACKER_MISSING_MESSAGE % ("report_operation",)
        get_one_time_logger().log_once(message, "report_operation", logging.WARNING)
        return

    tracker.log_dataset(op_report=operation_report)


def log_dataset_op(
    op_path: Union[Target, str],
    op_type: Union[DbndDatasetOperationType, str],
    *,
    success: bool = True,
    error: Optional[str] = None,
    data: Optional[Any] = None,
    with_histograms: Optional[Union[bool, str, List[str], LogDataRequest]] = None,
    with_partition: Optional[bool] = None,
    with_stats: Optional[Union[bool, str, List[str], LogDataRequest]] = True,
    with_preview: bool = False,
    with_schema: bool = True,
    send_metrics: bool = True,
    row_count: Optional[int] = None,
    column_count: Optional[int] = None,
    operation_source: str = LOG_DATASET_OP_OP_SOURCE,
):
    """
    Logs dataset operation and meta data to dbnd.

    Args:
        op_path: Target object to log or a unique path representing the operation location.
        op_type: Type of operation that been done with the target - read, write, delete.
        success: True if the operation succeeded, False otherwise.
        error: Optional error message.
        data: Optional value of data to use build meta-data on the target.
        with_histograms: Should calculate histogram of the given data - relevant only with data param.
            - Boolean to calculate or not on all the data columns.
        with_stats: Should extract schema of the data as meta-data of the target - relevant only with data param.
            - Boolean to calculate or not on all the data columns.
        with_partition: If True, the webserver tries to detect partitions of our datasets and extract them from the path,
                        otherwise not manipulating the dataset path at all.
        with_preview: Should extract preview of the data as meta-data of the target - relevant only with data param.
        with_schema: Should extract schema of the data as meta-data of the target - relevant only with data param.
        send_metrics: Should report preview, schemas and histograms as metrics.
        row_count: Should report row count no matter what is the data
        column_count: Should report column count no matter what is the data
        operation_source: Optional source of operation, such as: redshift_tracker, dbt_sdk, etc...
    Example::

        @task
        def prepare_data():
            log_dataset_op(
                "/path/to/value.csv",
                DbndDatasetOperationType.read,
                data=pandas_data_frame,
                with_preview=True,
                with_schema=True,
            )
    """
    operation_report = DatasetOperationReport(
        op_path=op_path,
        op_type=op_type,
        data=data,
        success=success,
        error=error,
        with_preview=with_preview,
        with_schema=with_schema,
        with_histograms=with_histograms,
        with_stats=with_stats,
        send_metrics=send_metrics,
        with_partition=with_partition,
        row_count=row_count,
        column_count=column_count,
        operation_source=operation_source,
    )
    _report_operation(operation_report)


@seven.contextlib.contextmanager
def dataset_op_logger(
    op_path: Union[Target, str],
    op_type: Union[DbndDatasetOperationType, str],
    *,
    data: Optional[Any] = None,
    with_histograms: Optional[bool] = None,
    with_partition: Optional[bool] = None,
    with_stats: Optional[bool] = True,
    with_preview: bool = False,
    with_schema: bool = True,
    send_metrics: bool = True,
    row_count: Optional[int] = None,
    column_count: Optional[int] = None,
    operation_source: str = LOG_DATASET_OP_OP_SOURCE,
):
    """
    Wrapper to Log dataset operation and meta data to dbnd.

    Make sure to only wrap operation related code!

    Args:
        op_path: Target object to log or a unique path representing the target logic location.
        op_type: Type of operation that been done with the dataset - read, write, delete.
        data: Optional value of data to use build meta-data on the dataset.
        with_histograms: Should calculate histogram of the given data - relevant only with data param.
            - Boolean to calculate or not on all the data columns.
        with_stats: Should extract schema of the data as meta-data of the target - relevant only with data param.
            - Boolean to calculate or not on all the data columns.
        with_partition: If True, the webserver tries to detect partitions of our datasets and extract them from the path,
                        otherwise not manipulating the dataset path at all.
        with_preview: Should extract preview of the data as meta-data of the dataset - relevant only with data param.
        with_schema: Should extract schema of the data as meta-data of the dataset - relevant only with data param.
        send_metrics: Should report preview, schemas and histograms as metrics.
        row_count: should report row count no matter what is the data
        column_count: should report column count no matter what is the data
        operation_source: Optional source of operation, such as: redshift tracker, dbt sdk, etc...
    Good Example::

        with dataset_op_logger("location://path/to/value.csv", "read"):
            value = read_from()
            # Read is successful

        unrelated_func()

    Bad Example::

        with dataset_op_logger("location://path/to/value.csv", "read"):
            value = read_from()
            # Read is successful
            unrelated_func()
            # If unrelated_func raises an exception, failed read operation is reported to databand.
    """
    operation_report = DatasetOperationReport(
        op_path=op_path,
        op_type=op_type,
        data=data,
        with_preview=with_preview,
        with_schema=with_schema,
        with_histograms=with_histograms,
        with_stats=with_stats,
        send_metrics=send_metrics,
        with_partition=with_partition,
        row_count=row_count,
        column_count=column_count,
        operation_source=operation_source,
    )

    try:
        yield operation_report
    except Exception as e:
        operation_report.set_error(error=str(e))
        raise
    else:
        operation_report.set_success()
    finally:
        _report_operation(operation_report)



# © Copyright Databand.ai, an IBM Company 2022

from six.moves import cPickle as pickle

import dbnd._vendor.cloudpickle as cloudpickle


# noinspection PyBroadException
def dumps(o):
    try:
        return pickle.dumps(o, pickle.HIGHEST_PROTOCOL)
    except Exception:
        return cloudpickle.dumps(o)


# noinspection PyBroadException
def loads(o):
    """Unpickle using cPickle"""
    return pickle.loads(o)



# © Copyright Databand.ai, an IBM Company 2022

from more_itertools import first

from dbnd._core.current import get_databand_context
from dbnd.api.query_params import build_query_api_params, create_filter_builder
from dbnd.api.shared_schemas.job_schema import JobSchemaV2


def get_job_by_name(job_name):
    query_params = build_query_api_params(filters=build_job_name_filter(job_name))

    job_result = _get_job(get_databand_context().databand_api_client, query_params)
    return first(job_result, None)


def is_job_exists(job_name):
    return bool(get_job_by_name(job_name))


def _get_job(client, query_api_args):
    endpoint = "?".join(["jobs", query_api_args])
    schema = JobSchemaV2(strict=False)
    response = client.api_request(endpoint=endpoint, data="", method="GET")
    return schema.load(data=response["data"], many=True).data


build_job_name_filter = create_filter_builder("name", "eq")



# © Copyright Databand.ai, an IBM Company 2022

import logging
import sys

from types import FunctionType, ModuleType

from dbnd import task
from dbnd._core.tracking.no_tracking import should_not_track


logger = logging.getLogger(__name__)


def _is_function(obj):
    return isinstance(obj, FunctionType)


def _is_task(obj):
    """
    checks if obj is decorated func (dbnd generated object)
    """
    return hasattr(obj, "__is_dbnd_task__")


def _track_function(function):
    if not _is_function(function) or should_not_track(function) or _is_task(function):
        return

    decorated_function = task(function)

    # We modify all modules since each module has its own pointers to local and imported functions.
    # If a module has already imported the function we need to change the pointer in that module.
    for module in sys.modules.copy().values():
        if not _is_module(module):
            continue

        for k, v in module.__dict__.items():
            if v is function:
                module.__dict__[k] = decorated_function


def track_functions(*args):
    """Track functions by decorating them with @task."""
    for arg in args:
        try:
            _track_function(arg)
        except Exception:
            logger.exception("Failed to track %s" % arg)


def _is_module(obj):
    return isinstance(obj, ModuleType)


def track_module_functions(module):
    """
    Track functions inside module by decorating them with @task.

    Only functions implemented in module will be tracked, imported functions won't be tracked.
    """
    try:
        if not _is_module(module):
            return

        module_objects = module.__dict__.values()
        module_functions = [i for i in module_objects if _is_module_function(i, module)]
        track_functions(*module_functions)
    except Exception:
        logger.exception("Failed to track %s" % module)


def track_modules(*args):
    """
    Track functions inside modules by decorating them with @task.

    Only functions implemented in module will be tracked, imported functions won't be tracked.
    """
    for arg in args:
        try:
            track_module_functions(arg)
        except Exception:
            logger.exception("Failed to track %s" % arg)


def _is_module_function(function, module):
    try:
        if not _is_function(function):
            return False

        if not hasattr(function, "__globals__"):
            return False

        return function.__globals__ is module.__dict__
    except Exception:
        logger.exception("Failed to track %s" % function)
        return False



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import print_function

import logging
import sys

from dbnd._core.current import try_get_databand_context
from dbnd._core.task_build.dbnd_decorator import task
from dbnd.tasks.doctor.doctor_report_builder import DoctorStatusReportBuilder


logger = logging.getLogger(__name__)


@task
def logging_status():
    # type: ()->str
    """
    Shows the status of the logging system
    All known loggers, logging configuration and so on.
    :return:
    """
    report = DoctorStatusReportBuilder("Logging Status")

    report.log("logging.root", logging.root)
    report.log("logging.root.handlers", logging.root.handlers)
    report.log("logger", logger)
    report.log("logger.handlers", logger.handlers)

    # airflow usually alternate stderr/stdout
    report.log("sys.stderr", sys.stderr)
    report.log("sys.stderr[close]", hasattr(sys.stderr, "close"))
    report.log("sys.stderr", sys.__stderr__)
    report.log("sys.__stderr__[close]", hasattr(sys.__stderr__, "close"))

    dbnd_context = try_get_databand_context()
    if dbnd_context:
        from dbnd._core.task_ctrl.task_visualiser import TaskVisualiser

        report.add_sub_report(
            TaskVisualiser(dbnd_context.settings.log).banner("Log Config")
        )
    # check airflow logging

    try:
        from logging import Logger

        airflow_task_logger = Logger.manager.loggerDict.get("airflow.task")
        if airflow_task_logger:
            report.log("Airlfow task logger", airflow_task_logger)
            report.log("Airlfow task logger handlers", airflow_task_logger.handlers)
        else:
            report.log("Airlfow task logger", "not found")
    except Exception as ex:
        ex_msg = "Failed to get airlfow.task logger status: %s" % ex
        report.log("Airflow task logger", ex_msg)
        logger.exception(ex_msg)

    logging_status = report.get_status_str()
    logging_status = "\n{sep}\n{msg}\n{sep}s\n".format(msg=logging_status, sep="*" * 40)
    logger.info(logging_status)
    # if we run this check we might have a problem with logs, we don't know how we are going to see the message
    print("\n\nLogging Status (via __stderr__)%s" % logging_status, file=sys.__stderr__)

    logger.info("Running logging validation.. (you will see a lot of messages)")

    # now we can print things, it might be that one of them will "kill the process"
    # because of some weird log handlers loop
    print("Message via print")
    print("Message via print stderr", file=sys.stderr)
    print("Message via print __stderr__", file=sys.__stderr__)
    logging.info("Message via logging root")
    logger.info("Message via logger")

    return logging_status



# © Copyright Databand.ai, an IBM Company 2022

import importlib
import logging
import os
import re
import sys

from dbnd._core.errors import DatabandError, friendly_error
from dbnd._core.utils.basics.memoized import cached


logger = logging.getLogger(__name__)

try:
    import_errors = (ImportError, ModuleNotFoundError)
except Exception:
    # we are python2
    import_errors = (ImportError,)


@cached()
def _load_module(module, description):
    try:
        try:
            return importlib.import_module(module)
        except import_errors:

            # in some cases it will not help
            # like "tests" package.
            # it too late to fix it as tests already loaded from site-packages..
            if os.getcwd() in sys.path:
                raise

            # we'll try to load current folder to PYTHONPATH, just in case
            logger.info(
                "Databand has failed to load module '%s', "
                "it will retry with cwd at PYTHONPATH." % module
            )
            sys.path.insert(0, os.getcwd())
            m = importlib.import_module(module)
            logger.info(
                "We have managed to load module after adding %s to PYTHONPATH, "
                "please consider using 'pip install -e . ' with your project"
                % os.getcwd()
            )
            return m
    except import_errors as ex:
        logger.warning(
            "Failed to load module '%s' %s: cwd='%s', sys.path=\n\t%s",
            module,
            friendly_error.dbnd_module_not_found_tip(module),
            os.getcwd(),
            "\n\t".join(sys.path),
        )
        raise friendly_error.failed_to_import_user_module(
            ex, module=module, description=description
        )


def load_python_module(module, module_source):
    logger.info("Loading modules '%s' from %s.", module, module_source)
    for m in module.split(","):
        _load_module(m, module_source)


def load_python_attr_from_module(attr_path):
    m = re.match(r"^(\S+)\.(\S+)", attr_path)
    if not m:
        raise friendly_error.config.wrong_func_attr_format(attr_path)

    module_path, attr_name = m.group(1), m.group(2)
    module = _load_module(module_path, description="")

    if not hasattr(module, attr_name):
        raise DatabandError("Failed to import symbol %s" % attr_path)

    attr = getattr(module, attr_name)
    return attr


def load_python_callable(callable_path):
    callable_attr = load_python_attr_from_module(callable_path)
    if not callable(callable_attr):
        raise DatabandError("The `%s` is not `callable`" % callable_attr)
    return callable_attr


def run_user_func(callable_path):
    if not callable_path:
        return None
    f = load_python_callable(callable_path=callable_path)
    try:
        return f()
    except Exception:
        logger.warning("Failed to run user function %s", callable_path)
        raise



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from dbnd._core.constants import TaskRunState
from dbnd._core.errors import DatabandRuntimeError
from dbnd._core.task_run.task_run_logging import TaskRunLogManager
from dbnd._core.task_run.task_run_meta_files import TaskRunMetaFiles
from dbnd._core.task_run.task_run_runner import TaskRunRunner
from dbnd._core.task_run.task_run_sync_local import TaskRunLocalSyncer
from dbnd._core.task_run.task_run_tracker import TaskRunTracker
from dbnd._core.task_run.task_sync_ctrl import TaskSyncCtrl
from dbnd._core.tracking.registry import get_tracking_store
from dbnd._core.utils.string_utils import clean_job_name
from dbnd._core.utils.timezone import utcnow
from dbnd._core.utils.uid_utils import get_task_run_attempt_uid_by_task_run, get_uuid
from targets import target


logger = logging.getLogger(__name__)
if typing.TYPE_CHECKING:
    from typing import Any

    from dbnd._core.run.databand_run import DatabandRun
    from dbnd._core.settings import EngineConfig
    from dbnd._core.task.task import Task


class TaskRun(object):
    def __init__(
        self,
        task,
        run,
        task_af_id=None,
        try_number=1,
        is_dynamic=None,
        task_engine=None,
    ):
        # type: (Task, DatabandRun, str, int, bool, EngineConfig)-> None
        # actually this is used as Task uid

        self.task = task  # type: Task
        self.run = run  # type: DatabandRun
        self.task_engine = task_engine
        self.is_dynamic = is_dynamic if is_dynamic is not None else task.task_is_dynamic
        self.is_system = task.task_is_system
        self.task_af_id = task_af_id or self.task.task_id

        if task.ctrl.force_task_run_uid:
            self.task_run_uid = tr_uid = task.ctrl.force_task_run_uid
            if isinstance(tr_uid, TaskRunUidGen):
                self.task_run_uid = tr_uid.generate_task_run_uid(
                    run=run, task=task, task_af_id=self.task_af_id
                )
        else:
            self.task_run_uid = get_uuid()

        # used by all kind of submission controllers
        self.job_name = clean_job_name(self.task_af_id).lower()
        self.job_id = self.job_name + "_" + str(self.task_run_uid)[:8]

        # custom per task engine , or just use one from global env
        dbnd_local_root = (
            self.task_engine.dbnd_local_root or self.run.env.dbnd_local_root
        )
        self.local_task_run_root = (
            dbnd_local_root.folder(run.run_folder_prefix)
            .folder("tasks")
            .folder(self.task.task_id)
        )

        self.attempt_number = try_number
        self.task_run_attempt_uid = None
        self.attempt_folder = None
        self.meta_files = None
        self.log = None
        self.init_new_task_run_attempt()

        # TODO: inherit from parent task if disabled
        self.is_tracked = task._conf__tracked

        if self.is_tracked and self.run.is_tracked:
            tracking_store = self.run.context.tracking_store
        else:
            tracking_store = get_tracking_store(
                self.run.context,
                tracking_store_names=["console"],
                api_channel_name=None,
                max_retires=1,
                tracker_raise_on_error=False,
                remove_failed_store=True,
            )

        self.tracking_store = tracking_store
        self.tracker = TaskRunTracker(task_run=self, tracking_store=tracking_store)
        self.runner = TaskRunRunner(task_run=self)
        self.deploy = TaskSyncCtrl(task_run=self)
        self.sync_local = TaskRunLocalSyncer(task_run=self)
        self.task_tracker_url = self.tracker.task_run_url()
        self.external_resource_urls = dict()
        self.errors = []

        self.is_root = False
        self.is_reused = False
        self.is_skipped = False
        # Task can be skipped as it's not required by any other task scheduled to run
        self.is_skipped_as_not_required = False

        self.airflow_context = None
        self._task_run_state = None

        self.start_time = None
        self.finished_time = None

    def __getstate__(self):
        d = self.__dict__.copy()
        if "airflow_context" in d:
            # airflow context contains "auto generated code" that can not be pickled (Vars class)
            # we don't need to pickle it as we pickle DAGs separately
            d = self.__dict__.copy()
            del d["airflow_context"]
        return d

    @property
    def task_run_env(self):
        return self.run.context.task_run_env

    def task_run_attempt_file(self, *path):
        return target(self.attempt_folder, *path)

    @property
    def last_error(self):
        return self.errors[-1] if self.errors else None

    def _get_log_files(self):

        log_local = None
        if self.log.local_log_file:
            log_local = self.log.local_log_file.path

        log_remote = None
        if self.log.remote_log_file:
            log_remote = self.log.remote_log_file.path

        return log_local, log_remote

    @property
    def task_run_state(self):
        return self._task_run_state

    @task_run_state.setter
    def task_run_state(self, value):
        raise AttributeError("Please use explicit .set_task_run_state()")

    def set_task_run_state(self, state, track=True, error=None):
        # type: (TaskRunState, bool, Any) -> bool
        # Optional bool track param - will send tracker.set_task_run_state() by default
        if not state or self._task_run_state == state:
            return False

        if error:
            self.errors.append(error)
        if state == TaskRunState.RUNNING:
            self.start_time = utcnow()

        self._task_run_state = state
        if track:
            self.tracking_store.set_task_run_state(
                task_run=self, state=state, error=error
            )
        return True

    def set_task_reused(self):
        self._task_run_state = TaskRunState.SUCCESS
        self.tracking_store.set_task_reused(task_run=self)

    def set_external_resource_urls(self, links_dict):
        # if value is None skip
        if links_dict is None:
            # TODO: Throw exception?
            return

        for link in links_dict:
            if not links_dict[link]:
                # Dict has empty fields
                # TODO: Throw exception?
                return

        self.external_resource_urls.update(links_dict)
        self.tracking_store.save_external_links(
            task_run=self, external_links_dict=links_dict
        )

    def update_task_run_attempt(self, attempt_number):
        if attempt_number is None:
            raise DatabandRuntimeError("cannot set None as the attempt number")

        if self.attempt_number != attempt_number:
            self.attempt_number = attempt_number
            self.init_new_task_run_attempt()

    def init_new_task_run_attempt(self):
        # trying to find if we should use attempt_uid that been set from external process.
        # if so - the attempt_uid is uniquely for this task_run_attempt, and that why we pop.
        self.task_run_attempt_uid = get_task_run_attempt_uid_by_task_run(self)

        self.attempt_folder = self.task._meta_output.folder(
            "attempt_%s_%s" % (self.attempt_number, self.task_run_attempt_uid),
            extension=None,
        )
        self.attempt_folder_local = self.local_task_run_root.folder(
            "attempt_%s_%s" % (self.attempt_number, self.task_run_attempt_uid),
            extension=None,
        )
        self.attemp_folder_local_cache = self.attempt_folder_local.folder("cache")
        self.meta_files = TaskRunMetaFiles(self.attempt_folder)
        self.log = TaskRunLogManager(task_run=self)

    def __repr__(self):
        return "TaskRun(id=%s, af_id=%s)" % (self.task.task_id, self.task_af_id)


class TaskRunUidGen(object):
    def generate_task_run_uid(self, run, task, task_af_id):
        return get_uuid()



# © Copyright Databand.ai, an IBM Company 2022

import enum
import logging

from typing import Any, Dict, Optional

import attr

from dbnd._core.parameter import PARAMETER_FACTORY as parameter
from dbnd._core.task import Config
from targets import Target
from targets.value_meta import _DEFAULT_VALUE_PREVIEW_MAX_LEN, ValueMeta, ValueMetaConf
from targets.values import (
    ObjectValueType,
    TargetValueType,
    ValueType,
    get_value_type_of_obj,
)


logger = logging.getLogger()


class ValueTrackingLevel(enum.Enum):
    """
    Multiple strategies with different limitations on potentially expensive calculation for value_meta
    """

    NONE = 1
    SMART = 2
    ALL = 3


class TrackingConfig(Config):
    _conf__task_family = "tracking"

    project = parameter(
        default=None,
        description="Project to which run should be assigned. "
        "If not set default project is used. Tracking server will select project with is_default == True.",
    )[str]

    databand_external_url = parameter(
        default=None,
        description="Tracker URL to be used for tracking from external systems",
    )[str]

    log_value_size = parameter(
        default=True,
        description="Calculate and log value size (can cause a full scan on not-indexable distributed memory objects) ",
    )[bool]

    log_value_schema = parameter(
        default=True, description="Calculate and log value schema "
    )[bool]

    log_value_stats = parameter(
        default=True,
        description="Calculate and log value stats(expensive to calculate, better use log_stats on parameter level)",
    )[bool]

    log_value_preview = parameter(
        default=True,
        description="Calculate and log value preview. Can be expensive on Spark.",
    )[bool]

    log_value_preview_max_len = parameter(
        description="Max size of value preview to be saved at DB, max value=50000"
    ).value(_DEFAULT_VALUE_PREVIEW_MAX_LEN)

    log_value_meta = parameter(
        default=True, description="Calculate and log value meta "
    )[bool]

    log_histograms = parameter(
        default=True,
        description="Enable calculation and tracking of histograms. Can be expensive",
    )[bool]

    value_reporting_strategy = parameter(
        default=ValueTrackingLevel.SMART,
        description="Multiple strategies with different limitations on potentially expensive calculation for value_meta."
        "ALL => no limitations."
        "SMART => restrictions on lazy evaluation types."
        "NONE (default) => limit everything.",
    ).enum(ValueTrackingLevel)

    track_source_code = parameter(
        default=False,
        description="Enable tracking of function, module and file source code",
    )[bool]

    auto_disable_slow_size = parameter(
        default=True,
        description="Auto disable slow preview for Spark DF with text formats",
    )[bool]

    flatten_operator_fields = parameter(
        default={},
        description="Control which of the operator's fields would be flatten when tracked",
    )[Dict[str, str]]

    capture_tracking_log = parameter(
        default=False, description="Enable log capturing for tracking tasks"
    )[bool]

    def get_value_meta_conf(self, meta_conf, value_type, target=None):
        # type: (ValueMetaConf, ValueType, Optional[Target]) -> ValueMetaConf
        meta_conf_by_type = calc_meta_conf_for_value_type(
            self.value_reporting_strategy, value_type, target
        )
        # translating TrackingConfig to meta_conf
        meta_conf_by_config = self._build_meta_conf()
        return meta_conf.merge_if_none(meta_conf_by_type).merge_if_none(
            meta_conf_by_config
        )

    def _build_meta_conf(self):
        # type: () -> ValueMetaConf
        """
        Translate this configuration into value meta conf
        WE EXPECT IT TO HAVE ALL THE INNER VALUES SET WITHOUT NONES
        """
        return ValueMetaConf(
            log_schema=self.log_value_schema,
            log_size=self.log_value_size,
            log_preview_size=self.log_value_preview_max_len,
            log_preview=self.log_value_preview,
            log_stats=self.log_value_stats,
            log_histograms=self.log_histograms,
        )


def _is_default_value_type(value_type):
    return value_type is None or isinstance(value_type, ObjectValueType)


def get_value_meta(value, meta_conf, tracking_config, value_type=None, target=None):
    # type: ( Any, ValueMetaConf, TrackingConfig, Optional[ValueType], Optional[Target]) -> Optional[ValueMeta]
    """
    Build the value meta for tracking logging.
    Using the given meta config, the value, and tracking_config to calculate the required value meta.

    @param value: the value to calc value meta for
    @param meta_conf: a given meta_config by a user
    @param tracking_config: TrackingConfig to calc the wanted meta conf
    @param value_type: optional value_type, if its known.
    @param target: knowledge about the target which contains the value - this can effect the cost of the calculation
    @return: Calculated value meta
    """

    if value is None:
        return None

    # required for calculating the relevant configuration and to build value_meta
    if _is_default_value_type(value_type) or isinstance(value_type, TargetValueType):
        # we calculate the actual value_type even if the given value is the default value
        # so we can be sure that we can report it the right way
        # also Targets are futures types and now would can log their actual value
        value_type = get_value_type_of_obj(value, default_value_type=ObjectValueType())

    meta_conf = tracking_config.get_value_meta_conf(meta_conf, value_type, target)
    return value_type.get_value_meta(value, meta_conf=meta_conf)


def calc_meta_conf_for_value_type(tracking_level, value_type, target=None):
    # type: (ValueTrackingLevel, ValueType, Optional[Target]) -> ValueMetaConf
    """
    Calculating the right value log config base on the value type in order control the tracking of
    lazy evaluated types like spark dataframes

    IMPORTANT - The result is ValueMetaConf with restrictions only! this should be merged into a full ValueMetaConf.
    """

    if tracking_level == ValueTrackingLevel.ALL:
        # no restrictions
        return ValueMetaConf()

    if tracking_level == ValueTrackingLevel.SMART:

        result = ValueMetaConf()
        if value_type.is_lazy_evaluated:
            # restrict only for lazy evaluate values
            result = ValueMetaConf.disabled_expensive()

        elif target is not None and not value_type.support_fast_count(target):
            # we don't set it to True cause there might
            # be different configuration that will want it to be False
            result = attr.evolve(result, log_size=False)

        return result

    if tracking_level == ValueTrackingLevel.NONE:
        # restrict any
        return ValueMetaConf.disabled_expensive()



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Dict, List

from dbnd._core.configuration.environ_config import in_tracking_mode
from dbnd._core.constants import CloudType
from dbnd._core.log import dbnd_log_debug
from dbnd._core.parameter import PARAMETER_FACTORY as parameter
from dbnd._core.task import Config


logger = logging.getLogger()


class DatabandSystemConfig(Config):
    """Databand's command line arguments (see `dbnd run --help`)"""

    _conf__task_family = "databand"

    verbose = parameter(description="Make logging output more verbose").value(False)
    print_task_band = parameter(description="Print task_band in logging output.").value(
        False
    )

    describe = parameter.help("Describe current run").value(False)

    module = parameter(
        default=None, description="Auto load this module before resolving user classes"
    )[str]
    env = parameter(
        default=CloudType.local,
        description="task environment: (based on core.environments) local/aws/aws_prod/gcp/prod",
    )[str]

    conf_file = parameter(default=None, description="List of files to read from")[
        List[str]
    ]
    # keep it last one, parameter keyword is used for parameters generation
    # all following params will be dict
    conf = parameter(
        default=None,
        description="JSON string/key=value that gets into the Task attribute",
    )[Dict[str, str]]


class CoreConfig(Config):
    """Databand's core functionality behaviour"""

    _conf__task_family = "core"

    databand_url = parameter(
        default=None,
        description="Tracker URL to be used for creating links in console logs",
    )[str]
    databand_access_token = parameter(
        description="Personal access token to connect to the dbnd web server",
        default=None,
        hidden=True,
    )[str]

    ignore_ssl_errors = parameter(
        description="Ignore SSL errors to allow usage of the self-signed certificate on dbnd web server",
        default=False,
    )[bool]

    extra_default_headers = parameter(
        description="Specify extra headers to be used as defaults for databand_api_client.",
        default=None,
    )[Dict[str, str]]

    tracker = parameter(
        default=["file", "console", "api"], description="Tracking Stores to be used"
    )[List[str]]

    tracker_api = parameter(
        default="web", description="Tracker Channels to be used by 'api' store"
    )[str]

    tracker_version = parameter[str]

    debug_webserver = parameter(
        description="Allow collecting the webservers logs for each api-call on the local machine. "
        "Requires that the web-server supports and allow this.",
        default=False,
    )[bool]

    silence_tracking_mode = parameter(
        default=False, description="Silence console on tracking mode"
    )[bool]

    tracker_raise_on_error = parameter(
        default=True, description="Raise error when failed to track data"
    )[bool]
    remove_failed_store = parameter(
        default=False, description="Remove a tracking store if it fails"
    )[bool]

    max_tracking_store_retries = parameter(
        default=2,
        description="Maximal amounts of retries allowed for a single tracking store call if it failed.",
    )[int]

    client_session_timeout = parameter(
        description="Minutes to recreate the api client's session", default=5
    )[int]
    client_max_retry = parameter(
        description="Maximum amount of retries on failed connection for the api client",
        default=2,
    )[int]
    client_retry_sleep = parameter(
        description="Sleep between retries of the api client", default=0.1
    )[float]

    # USER CODE TO RUN ON START
    user_configs = parameter(
        empty_default=True,
        description="Contains the config for creating tasks from user code",
    )[List[str]]

    # user_pre_init = defined at Databand System config, dbnd_on_pre_init_context
    user_init = parameter(
        default=None,
        description="Runs in every DBND process with System configuration in place (dbnd_post_enter_context)",
    )[object]
    user_driver_init = parameter(
        default=None,
        description="Runs in driver after config initialization (dbnd_on_new_context)",
    )[object]

    user_code_on_fork = parameter(
        default=None, description="Runs in sub process (parallel/kubernetes/external)"
    )[object]

    # PLUGINS
    plugins = parameter(
        description="plugins to load on databand context creation", default=None
    )[str]
    fix_env_on_osx = parameter(
        description="add no_proxy=* to env vars, fixing issues with multiprocessing on osx"
    )[bool]

    environments = parameter(description="List of enabled environments")[List[str]]

    #### TO DEPRECATE ## (DONT USE)
    # deprecate in favor of databand_access_token
    dbnd_user = parameter(
        description="DEPRECATED: user used to connect to the dbnd web server"
    )[str]
    dbnd_password = parameter(
        description="DEPRECATED: password used to connect to the dbnd web server",
        hidden=True,
    )[str]

    # deprecated at 0.34 (Backward compatibility)
    tracker_url = parameter(
        default=None,
        description="DEPRECATED: Tracker URL to be used for creating links in console logs",
    )[str]

    def _validate(self):
        if not self.databand_url and self.tracker_url:
            logger.warning(
                "core.databand_url was not set, using deprecated 'core.tracker_url' instead."
            )
            self.databand_url = self.tracker_url

        if self.databand_url and self.databand_url.endswith("/"):
            dbnd_log_debug(
                "Please fix your core.databand_url value, "
                "it should not contain '/' at the end, auto-fix has been applied."
            )
            self.databand_url = self.databand_url[:-1]

        # automatically disabling tracking if databand_url is not set
        if not self.databand_url:
            dbnd_log_debug(
                "Automatically disabling tracking to databand service as databand_url is not set"
            )
            self.tracker = [t for t in self.tracker if t != "api"]

        if self.databand_access_token and (self.dbnd_user or self.dbnd_password):
            logger.debug(
                "core.databand_access_token is used instead of defined dbnd_user and dbnd_password."
            )

        if not (in_tracking_mode() and "api" in self.tracker):
            self.silence_tracking_mode = False

        # in tracking mode we track to console only if we are not tracking to the api
        if self.silence_tracking_mode:
            self.tracker = [t for t in self.tracker if t != "console"]

    def build_tracking_store(self, databand_ctx, remove_failed_store=None):
        from dbnd._core.tracking.registry import get_tracking_store

        if remove_failed_store is None:
            remove_failed_store = self.remove_failed_store

        return get_tracking_store(
            databand_ctx,
            tracking_store_names=self.tracker,
            api_channel_name=self.tracker_api,
            max_retires=self.max_tracking_store_retries,
            tracker_raise_on_error=self.tracker_raise_on_error,
            remove_failed_store=remove_failed_store,
        )

    def build_databand_api_client(self):
        from dbnd.utils.api_client import ApiClient

        credentials = (
            {"token": self.databand_access_token}
            if self.databand_access_token
            else {"username": self.dbnd_user, "password": self.dbnd_password}
        )

        return ApiClient(
            self.databand_url,
            credentials=credentials,
            debug_server=self.debug_webserver,
            session_timeout=self.client_session_timeout,
            default_max_retry=self.client_max_retry,
            default_retry_sleep=self.client_retry_sleep,
            extra_default_headers=self.extra_default_headers,
            ignore_ssl_errors=self.ignore_ssl_errors,
        )



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Any, Dict

import six

from dbnd._core.current import in_tracking_run, is_orchestration_run
from dbnd._core.errors.base import (
    DatabandAuthenticationError,
    DatabandWebserverNotReachableError,
)
from dbnd._core.errors.errors_utils import log_exception
from dbnd._core.tracking.backends import TrackingStore, TrackingStoreThroughChannel
from dbnd._core.tracking.backends.abstract_tracking_store import is_state_call


MAX_RETRIES = 2

logger = logging.getLogger(__name__)


def try_run_handler(tries, store, handler_name, kwargs):
    # type: (int, TrackingStore, str, Dict[str, Any]) -> Any
    """
    Locate the handler function to run and will try to run it multiple times.
    If fails all the times -> raise the last error.

    @param tries: maximum amount of retries, positive integer.
    @param store: the store to run its handler
    @param handler_name: the name of the handler to run
    @param kwargs: the input for the handler
    @return: The result of the handler if succeeded, otherwise raise the last error
    """
    try_num = 1
    handler = getattr(store, handler_name)

    while True:
        try:
            return handler(**kwargs)

        except DatabandAuthenticationError as ex:
            # No need to retry with the same credentials, log exception and exit immediately.
            log_exception(
                f"Auth Failed storing tracking information from {handler_name} at {str(store)}",
                ex,
                non_critical=True,
            )
            # Failed auth is also not a reson to stop attempting to track subsequent request.
            # We have an option to process requests with failed auth on demand.
            # So, fake success:
            return {}
        except Exception as ex:
            log_exception(
                f"Try {try_num}/{tries}: Tracking failed from {handler_name} at {str(store)}",
                ex,
                non_critical=True,
            )

            if try_num == tries:
                # raise on the last try
                raise

            try_num += 1


def build_store_name(name, channel_name):
    if channel_name:
        return "{}-{}".format(name, channel_name)

    return name


class CompositeTrackingStore(TrackingStore):
    def __init__(
        self,
        tracking_stores,
        max_retires,
        raise_on_error=True,
        remove_failed_store=False,
    ):
        # type: (Dict[str, TrackingStore], int, bool, bool) -> CompositeTrackingStore

        if not tracking_stores:
            logger.warning("You are running without any tracking store configured.")

        self._stores = tracking_stores
        self._raise_on_error = raise_on_error
        self._remove_failed_store = remove_failed_store
        self._max_retries = max_retires

    def _invoke(self, name, kwargs):
        res = None
        failed_stores = []

        for store_name, store in six.iteritems(self._stores):
            tries = self._max_retries if is_state_call(name) else 1

            try:
                res = try_run_handler(tries, store, name, kwargs)
            except Exception as e:
                if self._remove_failed_store or (
                    in_tracking_run() and is_state_call(name)
                ):
                    failed_stores.append(store_name)

                if isinstance(e, DatabandWebserverNotReachableError):
                    if in_tracking_run():
                        logger.warning(str(e))

                    if is_orchestration_run():
                        # in orchestration runs we have good error collection that's show error banner
                        # error should have good msg and no need to show full trace
                        e.show_exc_info = False

                    if self._raise_on_error:
                        raise e

        if failed_stores:
            for store_name in failed_stores:
                logger.warning(
                    "Removing store {store_name}: {store} from stores list due to failure".format(
                        store_name=store_name, store=str(self._stores.get(store_name))
                    )
                )
                store = self._stores.pop(store_name)
                try:
                    store.flush()
                except Exception:
                    logger.exception(
                        f"Error during flush of {store_name} tracking backend"
                    )

            if not self._stores:
                logger.warning("You are running without any tracking store configured.")

        return res

    # this is a function that used for disabling Tracking api on spark inline tasks.
    def disable_tracking_api(self):
        filtered_stores = []

        for store in self._stores:
            if isinstance(store, TrackingStoreThroughChannel):
                continue
            filtered_stores.append(store)
        self._stores = filtered_stores

    def has_tracking_store(self, name, channel_name=None):
        name = build_store_name(name, channel_name)
        return name in self._stores

    @property
    def trackers_names(self):
        return list(self._stores.keys())

    def init_scheduled_job(self, **kwargs):
        return self._invoke(CompositeTrackingStore.init_scheduled_job.__name__, kwargs)

    def init_run(self, **kwargs):
        return self._invoke(CompositeTrackingStore.init_run.__name__, kwargs)

    def init_run_from_args(self, **kwargs):
        return self._invoke(CompositeTrackingStore.init_run_from_args.__name__, kwargs)

    def set_run_state(self, **kwargs):
        return self._invoke(CompositeTrackingStore.set_run_state.__name__, kwargs)

    def set_task_reused(self, **kwargs):
        return self._invoke(CompositeTrackingStore.set_task_reused.__name__, kwargs)

    def set_task_run_state(self, **kwargs):
        return self._invoke(CompositeTrackingStore.set_task_run_state.__name__, kwargs)

    def set_task_run_states(self, **kwargs):
        return self._invoke(CompositeTrackingStore.set_task_run_states.__name__, kwargs)

    def set_unfinished_tasks_state(self, **kwargs):
        return self._invoke(
            CompositeTrackingStore.set_unfinished_tasks_state.__name__, kwargs
        )

    def save_task_run_log(self, **kwargs):
        return self._invoke(CompositeTrackingStore.save_task_run_log.__name__, kwargs)

    def save_external_links(self, **kwargs):
        return self._invoke(CompositeTrackingStore.save_external_links.__name__, kwargs)

    def log_dataset(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_dataset.__name__, kwargs)

    def log_datasets(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_datasets.__name__, kwargs)

    def log_target(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_target.__name__, kwargs)

    def log_targets(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_targets.__name__, kwargs)

    def log_histograms(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_histograms.__name__, kwargs)

    def log_metrics(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_metrics.__name__, kwargs)

    def log_artifact(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_artifact.__name__, kwargs)

    def log_dbt_metadata(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_dbt_metadata.__name__, kwargs)

    def close(self):
        pass

    def add_task_runs(self, **kwargs):
        return self._invoke(CompositeTrackingStore.add_task_runs.__name__, kwargs)

    def heartbeat(self, **kwargs):
        return self._invoke(CompositeTrackingStore.heartbeat.__name__, kwargs)

    def save_airflow_task_infos(self, **kwargs):
        return self._invoke(
            CompositeTrackingStore.save_airflow_task_infos.__name__, kwargs
        )

    def update_task_run_attempts(self, **kwargs):
        return self._invoke(
            CompositeTrackingStore.update_task_run_attempts.__name__, kwargs
        )

    def is_ready(self, **kwargs):
        return all(store.is_ready() for store in self._stores.values())

    def flush(self):
        failed = False

        for name, store in self._stores.items():
            try:
                store.flush()
            except Exception as exc:
                failed = exc
                logger.exception(f"Error during flush of {name} tracking backend")

        if failed and self._raise_on_error:
            raise failed



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.utils.basics.singleton_context import SingletonContext


logger = logging.getLogger(__name__)


class OneTimeLogger(SingletonContext):
    """
    This class helps log messages for different types and makes sure that every type will be logged only once
    """

    def __init__(self):
        self._log_hostory = set()

    def log_once(self, message, message_type, log_level):
        if message_type not in self._log_hostory:
            logger.log(level=log_level, msg=message)
            self._log_hostory.add(message_type)


_OneTimeLogger = OneTimeLogger.try_instance()


def get_one_time_logger():
    return OneTimeLogger.get_instance()



# © Copyright Databand.ai, an IBM Company 2022

import typing

from dbnd._core.task_ctrl.task_ctrl import TaskSubCtrl


if typing.TYPE_CHECKING:
    from dbnd._core.task_run.task_run import TaskRun


class TaskRunCtrl(TaskSubCtrl):
    def __init__(self, task_run):
        # type: (TaskRun)-> None
        super(TaskRunCtrl, self).__init__(task=task_run.task)
        self.run = task_run.run
        self.task_run = task_run

        self.job = task_run  # backward compatible to old code

    @property
    def context(self):
        return self.run.context

    @property
    def task_run_uid(self):
        return self.task_run.task_run_uid

    @property
    def task_run_attempt_uid(self):
        return self.task_run.task_run_attempt_uid


class TaskJobCtrl(TaskRunCtrl):
    """
    Exists for backward compatabiltiy
    Please, do not use! It will be removed!
    """

    def __init__(self, task, job):
        super(TaskJobCtrl, self).__init__(task_run=job)
        self.job = job  # type: TaskRun



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.parameter.parameter_builder import (  # noqa: 401
    PARAMETER_FACTORY,
    ParameterFactory,
    build_parameter,
)
from dbnd._core.parameter.parameter_definition import infer_parameter_value_type
from dbnd._core.parameter.parameter_value import ParameterValue
from dbnd._core.parameter.parameters_mapper import ParametersMapper
from dbnd._core.utils.basics.nothing import NOTHING
from targets.values import (
    get_types_registry,
    get_value_type_of_obj,
    get_value_type_of_type,
)
from targets.values.builtins_values import DefaultObjectValueType


_PARAMS_MAPPER = ParametersMapper()  # type: ParametersMapper


def get_params_mapper():
    return _PARAMS_MAPPER


def register_custom_parameter(value_type, parameter):
    value_type = get_value_type_of_type(value_type, inline_value_type=True)

    _PARAMS_MAPPER.register_custom_parameter(value_type, parameter)
    get_types_registry().register_value_type(value_type)

    return parameter


def get_parameter_for_value_type(value_type):
    return _PARAMS_MAPPER.get_parameter(value_type)


def build_user_parameter_value(name, value, source):
    """
    Build parameter value for user defined name and value
    """
    value_type = get_value_type_of_obj(value, default_value_type=DefaultObjectValueType)
    param_f = get_parameter_for_value_type(value_type)
    param = build_parameter(param_f)
    param.name = name

    if value is NOTHING:
        parameter, warnings = param, []
        actual_value = param.default
    else:
        parameter, warnings = infer_parameter_value_type(param, value)
        actual_value = value

    return ParameterValue(
        parameter=parameter,
        source=source,
        source_value=value,
        value=actual_value,
        parsed=False,
        warnings=warnings,
    )



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import signal
import warnings

import dbnd

from dbnd._core.configuration.environ_config import (
    _env_banner,
    get_dbnd_project_config,
    is_unit_test_mode,
)
from dbnd._core.plugin.dbnd_plugins_mng import (
    register_dbnd_plugins,
    register_dbnd_user_plugins,
)
from dbnd._core.utils.basics.signal_utils import safe_signal
from dbnd._core.utils.platform import windows_compatible_mode
from dbnd._core.utils.platform.osx_compatible.requests_in_forked_process import (
    enable_osx_forked_request_calls,
)


logger = logging.getLogger(__name__)


def _surpress_loggers():
    logging.getLogger("googleapiclient.discovery_cache").setLevel(logging.ERROR)
    logging.getLogger("googleapiclient").setLevel(logging.WARN)


def _suppress_warnings():
    warnings.simplefilter("ignore", FutureWarning)


def _dbnd_exception_handling():
    if windows_compatible_mode:
        return

    # Enables graceful shutdown when running inside docker/kubernetes and subprocess shutdown
    # By default the kill signal is SIGTERM while our code mostly expects SIGINT (KeyboardInterrupt)
    def sigterm_handler(sig, frame):
        pid = os.getpid()
        logger.info("Received signal in default signal handler. PID: %s", pid)
        os.kill(pid, signal.SIGINT)

    safe_signal(signal.SIGTERM, sigterm_handler)


_dbnd_system_bootstrap = False


def dbnd_system_bootstrap():
    global _dbnd_system_bootstrap
    if _dbnd_system_bootstrap:
        return
    try:
        _dbnd_system_bootstrap = True

        # this will also initialize env if it's not initialized
        project_config = get_dbnd_project_config()
        if not project_config.quiet_mode:
            logger.info("Starting Databand %s!\n%s", dbnd.__version__, _env_banner())
            dbnd_run_info_source_version = os.environ.get(
                "DBND__RUN_INFO__SOURCE_VERSION"
            )
            if dbnd_run_info_source_version:
                logger.info("revision: %s", dbnd_run_info_source_version)
        from dbnd import dbnd_config

        dbnd_config.load_system_configs()
    except Exception:
        _dbnd_system_bootstrap = False
        raise


def fix_pyspark_imports():
    import sys

    pyspark_libs, regular_libs = [], []
    for p in sys.path:
        if "spark-core" in p or "pyspark.zip" in p:
            pyspark_libs.append(p)
        else:
            regular_libs.append(p)
    regular_libs.extend(pyspark_libs)
    sys.path = regular_libs


_dbnd_bootstrap = False
_dbnd_bootstrap_started = False


def dbnd_bootstrap():
    """Runs dbnd bootstrapping."""
    global _dbnd_bootstrap
    global _dbnd_bootstrap_started
    if _dbnd_bootstrap_started:
        return
    _dbnd_bootstrap_started = True

    dbnd_system_bootstrap()

    dbnd_project_config = get_dbnd_project_config()

    from targets.marshalling import register_basic_data_marshallers

    register_basic_data_marshallers()

    _surpress_loggers()
    _suppress_warnings()
    enable_osx_forked_request_calls()

    register_dbnd_plugins()

    from dbnd._core.configuration import environ_config
    from dbnd._core.configuration.dbnd_config import config
    from dbnd._core.plugin.dbnd_plugins import pm
    from dbnd._core.utils.basics.load_python_module import run_user_func

    user_plugins = config.get("core", "plugins", None)
    if user_plugins:
        register_dbnd_user_plugins(user_plugins.split(","))

    if is_unit_test_mode():
        pm.hook.dbnd_setup_unittest()

    pm.hook.dbnd_setup_plugin()

    if dbnd_project_config.is_sigquit_handler_on:
        from dbnd._core.utils.basics.signal_utils import (
            register_sigquit_stack_dump_handler,
        )

        register_sigquit_stack_dump_handler()

    # now we can run user code ( at driver/task)
    user_preinit = environ_config.get_user_preinit()
    if user_preinit:
        run_user_func(user_preinit)

    # if for any reason there will be code that calls dbnd_bootstrap, this will prevent endless recursion
    _dbnd_bootstrap = True



# COMMIT 1c62db5
"""
Vendorized from https://github.com/justanr/marshmallow_enum/releases/tag/v1.5.1
The MIT License (MIT)

Copyright (c) 2015 Alec Nikolas Reiter

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import unicode_literals

import sys
import warnings

from enum import Enum

from dbnd._vendor.marshmallow import ValidationError, fields


PY2 = sys.version_info.major == 2
# ugh Python 2
if PY2:
    string_types = (str, unicode)  # noqa: F821
    text_type = unicode  # noqa: F821
else:
    string_types = (str,)
    text_type = str


class LoadDumpOptions(Enum):
    value = 1
    name = 0


class EnumField(fields.Field):
    VALUE = LoadDumpOptions.value
    NAME = LoadDumpOptions.name

    default_error_messages = {
        "by_name": "Invalid enum member {input}",
        "by_value": "Invalid enum value {input}",
        "must_be_string": "Enum name must be string",
    }

    def __init__(
        self,
        enum,
        by_value=False,
        load_by=None,
        dump_by=None,
        error="",
        *args,
        **kwargs
    ):
        self.enum = enum
        self.by_value = by_value

        if error and any(old in error for old in ("{name", "{value", "{choices")):
            warnings.warn(
                "'name', 'value', and 'choices' fail inputs are deprecated,"
                "use input, names and values instead",
                DeprecationWarning,
                stacklevel=2,
            )

        self.error = error

        if load_by is None:
            load_by = LoadDumpOptions.value if by_value else LoadDumpOptions.name

        if load_by not in LoadDumpOptions:
            raise ValueError(
                "Invalid selection for load_by must be EnumField.VALUE or EnumField.NAME, got {}".format(
                    load_by
                )
            )

        if dump_by is None:
            dump_by = LoadDumpOptions.value if by_value else LoadDumpOptions.name

        if dump_by not in LoadDumpOptions:
            raise ValueError(
                "Invalid selection for load_by must be EnumField.VALUE or EnumField.NAME, got {}".format(
                    dump_by
                )
            )

        self.load_by = load_by
        self.dump_by = dump_by

        super(EnumField, self).__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        elif self.dump_by == LoadDumpOptions.value:
            return value.value
        else:
            return value.name

    def _deserialize(self, value, attr, data):
        if value is None:
            return None
        elif self.load_by == LoadDumpOptions.value:
            return self._deserialize_by_value(value, attr, data)
        else:
            return self._deserialize_by_name(value, attr, data)

    def _deserialize_by_value(self, value, attr, data):
        try:
            return self.enum(value)
        except ValueError:
            self.fail("by_value", input=value, value=value)

    def _deserialize_by_name(self, value, attr, data):
        if not isinstance(value, string_types):
            self.fail("must_be_string", input=value, name=value)

        try:
            return getattr(self.enum, value)
        except AttributeError:
            self.fail("by_name", input=value, name=value)

    def fail(self, key, **kwargs):
        kwargs["values"] = ", ".join([text_type(mem.value) for mem in self.enum])
        kwargs["names"] = ", ".join([mem.name for mem in self.enum])

        if self.error:
            if self.by_value:
                kwargs["choices"] = kwargs["values"]
            else:
                kwargs["choices"] = kwargs["names"]
            msg = self.error.format(**kwargs)
            raise ValidationError(msg)
        else:
            super(EnumField, self).fail(key, **kwargs)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from abc import ABCMeta

import six


logger = logging.getLogger(__name__)


@six.add_metaclass(ABCMeta)
class TrackingChannel(object):
    def _handle(self, name, data):
        logger.info("Tracking %s.%s is not implemented", self.__class__.__name__, name)

    def init_scheduled_job(self, data):
        return self._handle(TrackingChannel.init_scheduled_job.__name__, data)

    def init_run(self, data):
        return self._handle(TrackingChannel.init_run.__name__, data)

    def add_task_runs(self, data):
        return self._handle(TrackingChannel.add_task_runs.__name__, data)

    def set_run_state(self, data):
        return self._handle(TrackingChannel.set_run_state.__name__, data)

    def set_task_reused(self, data):
        return self._handle(TrackingChannel.set_task_reused.__name__, data)

    def update_task_run_attempts(self, data):
        return self._handle(TrackingChannel.update_task_run_attempts.__name__, data)

    def set_unfinished_tasks_state(self, data):
        return self._handle(TrackingChannel.set_unfinished_tasks_state.__name__, data)

    def save_task_run_log(self, data):
        return self._handle(TrackingChannel.save_task_run_log.__name__, data)

    def save_external_links(self, data):
        return self._handle(TrackingChannel.save_external_links.__name__, data)

    def log_dataset(self, data):
        return self._handle(TrackingChannel.log_dataset.__name__, data)

    def log_datasets(self, data):
        return self._handle(TrackingChannel.log_datasets.__name__, data)

    def log_target(self, data):
        return self._handle(TrackingChannel.log_target.__name__, data)

    def log_targets(self, data):
        return self._handle(TrackingChannel.log_targets.__name__, data)

    def log_metric(self, data):
        return self._handle(TrackingChannel.log_metric.__name__, data)

    def log_metrics(self, data):
        return self._handle(TrackingChannel.log_metrics.__name__, data)

    def log_artifact(self, data):
        return self._handle(TrackingChannel.log_artifact.__name__, data)

    def log_dbt_metadata(self, data):
        return self._handle(TrackingChannel.log_dbt_metadata.__name__, data)

    def heartbeat(self, data):
        return self._handle(TrackingChannel.heartbeat.__name__, data)

    def save_airflow_task_infos(self, data):
        return self._handle(TrackingChannel.save_airflow_task_infos.__name__, data)

    def is_ready(self):
        return self._handle(TrackingChannel.is_ready.__name__, None)

    def flush(self):
        pass



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import config


class SchedulerConfig(config.Config):
    """(Advanced) Databand's scheduler"""

    _conf__task_family = "scheduler"

    config_file = parameter(
        default=None, description="path to a file defining scheduled jobs to execute"
    )[str]

    # by default the scheduler drop-in will decide whether to sync or not based on if it's running inside the scheduler or not (and not the websrver)
    # the next two params can be used to force it one way or the other
    never_file_sync = parameter(
        default=False, description="disable syncing the scheduler config_file to the db"
    )[bool]

    always_file_sync = parameter(
        default=False, description="force syncing the scheduler config_file to the db"
    )[bool]

    no_ui_cli_edit = parameter(
        default=False,
        description="disables creating, editing and deleting scheduled jobs from the cli and ui. Scheduled job definitions will"
        "only be taken from the scheduler config file",
    )

    refresh_interval = parameter(
        default=1,
        description="interval to refresh the scheduled job list (from both the db and/or config file)",
    )[int]

    active_by_default = parameter(
        default=True, description="whether new scheduled jobs will be active by default"
    )[bool]

    default_retries = parameter(
        description="number of times to retry a failed run, unless set to a different value on the scheduled job"
    )[int]

    shell_cmd = parameter(
        description="If shell_cmd is True, the specified command will be executed through the shell. "
        "This can be useful if you are using Python primarily "
        "for the enhanced control flow it offers "
        "over most system shells and still want convenient access to other shell features "
        "such as shell pipes, filename wildcards, environment variable expansion, "
        "and expansion of ~ to a user's home directory."
    )[bool]



# © Copyright Databand.ai, an IBM Company 2022

import errno
import io
import logging
import os
import random
import shutil

from targets.errors import FileAlreadyExists, MissingParentDirectory, NotADirectory
from targets.fs.file_system import FileSystem
from targets.pipes.base import FileWrapper
from targets.utils.atomic import AtomicLocalFile


logger = logging.getLogger(__name__)


class LocalFileSystem(FileSystem):
    """
    Wrapper for access to file system operations.

    Work in progress - add things as needed.
    """

    name = "local"
    support_direct_access = True
    local = True

    def copy(self, old_path, new_path, raise_if_exists=False):
        if raise_if_exists and os.path.exists(new_path):
            raise RuntimeError("Destination exists: %s" % new_path)
        if os.path.isfile(old_path):
            d = os.path.dirname(new_path)
            if d and not os.path.exists(d):
                self.mkdir(d)
            shutil.copy(old_path, new_path)
        else:
            shutil.copytree(old_path, new_path)

    def exists(self, path):
        return os.path.exists(path)

    def mkdir(self, path, parents=True, raise_if_exists=False):
        if self.exists(path):
            if raise_if_exists:
                raise FileAlreadyExists()
            elif not self.isdir(path):
                raise NotADirectory()
            else:
                return

        if parents:
            try:
                os.makedirs(path)
            except OSError as err:
                # somebody already created the path
                if err.errno != errno.EEXIST:
                    raise
        else:
            if not os.path.exists(os.path.dirname(path)):
                raise MissingParentDirectory()
            os.mkdir(path)

    def mkdir_parent(self, path):
        """
        Create all parent folders if they do not exist.
        """
        parentfolder = os.path.dirname(os.path.normpath(path))
        if parentfolder:
            return self.mkdir(parentfolder)

    def isdir(self, path):
        return os.path.isdir(path)

    def listdir(self, path):
        for dir_, _, files in os.walk(path):
            assert dir_.startswith(path)
            for name in files:
                yield os.path.join(dir_, name)

    def remove(self, path, recursive=True):
        if recursive and self.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)

    def move(self, old_path, new_path, raise_if_exists=False):
        """
        Move file atomically. If source and destination are located
        on different filesystems, atomicity is approximated
        but cannot be guaranteed.
        """
        if raise_if_exists and os.path.exists(new_path):
            raise FileAlreadyExists("Destination exists: %s" % new_path)
        d = os.path.dirname(new_path)
        if d and not os.path.exists(d):
            self.mkdir(d)
        try:
            os.rename(old_path, new_path)
        except OSError as err:
            if err.errno == errno.EXDEV:
                new_path_tmp = "%s-%09d" % (new_path, random.randint(0, 999999999))
                shutil.copy(old_path, new_path_tmp)
                os.rename(new_path_tmp, new_path)
                os.remove(old_path)
            elif err.errno == errno.EEXIST:
                shutil.move(old_path, new_path)
            else:
                raise err

    def copy_from_local(self, local_path, dest, **kwargs):
        self.copy(local_path, dest, **kwargs)

    def move_from_local(self, local_path, dest, **kwargs):
        self.move(local_path, dest)

    def rename_dont_move(self, path, dest):
        """
        Rename ``path`` to ``dest``, but don't move it into the ``dest``
        folder (if it is a folder). This method is just a wrapper around the
        ``move`` method of LocalTarget.
        """
        self.move(path, dest, raise_if_exists=True)

    def open_read(self, path, mode="r"):
        file_io = io.FileIO(path, mode)
        return FileWrapper(io.BufferedReader(file_io))

    def open_write(self, path, mode="w", **kwargs):
        return AtomicLocalFile(path, fs=self, mode=mode, **kwargs)

    def move_dir(self, path, dest, raise_if_exists=False):
        if not self.exists(dest):
            self.mkdir(dest)

        elif not self.isdir(dest):
            raise NotADirectory(dest)

        for file in self.listdir(path):
            old_path = os.path.join(path, file)
            new_path = os.path.join(dest, file)
            self.move(old_path, new_path, raise_if_exists)

    def make_tmp(self):
        from tempfile import mkstemp

        _, tmp = mkstemp()
        return tmp

    def make_tmp_dir(self):
        from tempfile import mkdtemp

        return mkdtemp()



# © Copyright Databand.ai, an IBM Company 2022

import functools
import inspect
import logging
import typing

from typing import Any, Optional, Type

from dbnd._core.configuration.environ_config import get_dbnd_project_config
from dbnd._core.current import current_task_run, get_databand_run, try_get_current_task
from dbnd._core.errors import show_exc_info
from dbnd._core.errors.errors_utils import user_side_code
from dbnd._core.failures import dbnd_handle_errors
from dbnd._core.plugin.dbnd_airflow_operator_plugin import (
    build_task_at_airflow_dag_context,
    is_in_airflow_dag_build_context,
)
from dbnd._core.task.decorated_callable_task import _DecoratedCallableTask
from dbnd._core.task_build.task_context import TaskContextPhase, current_phase
from dbnd._core.task_build.task_metaclass import TaskMetaclass
from dbnd._core.task_build.task_passport import TaskPassport
from dbnd._core.tracking.managers.callable_tracking import CallableTrackingManager
from dbnd._core.utils.basics.nothing import NOTHING
from dbnd._core.utils.callable_spec import (
    CallableSpec,
    args_to_kwargs,
    build_callable_spec,
)
from dbnd._core.utils.lazy_property_proxy import CallableLazyObjectProxy
from dbnd._core.utils.one_time_logger import get_one_time_logger
from targets.inline_target import InlineTarget


logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from dbnd import Task
    from dbnd._core.task_run.task_run import TaskRun


class TaskDecorator(object):
    """
    This object represent the state and logic of decorated callable (user class or user function)
    All "expensive" objects are lazy: task_cls, callable_spec, callable_tracking_manager

    all "user-side" calls are routed into self.handle_callable_call() that will decide should we
    1. call user code directly (dbnd is disabled)
    2. call and track user callable call (tracking is enabled)
    3. create a Task that represents user code ( orchestration mode at @pipeline.band
    4. create a Task and run it (orchestration mode at @task.run)
    """

    def __init__(self, class_or_func, decorator_kwargs):
        # known parameters for @task
        self.class_or_func = class_or_func
        self.original_class_or_func = class_or_func

        self.task_type = decorator_kwargs.pop(
            "_task_type"
        )  # type: Type[_DecoratedCallableTask]
        self.task_default_result = decorator_kwargs.pop(
            "_task_default_result"
        )  # ParameterFactory
        self.task_defaults = decorator_kwargs.pop("defaults", None)

        self.task_namespace = decorator_kwargs.get("task_namespace", NOTHING)
        self.task_family = decorator_kwargs.get("_conf__task_family")

        # rest of kwargs are "user params"
        self.decorator_kwargs = decorator_kwargs

        self.task_passport = TaskPassport.build_task_passport(
            cls_name=self.original_class_or_func.__name__,
            module_name=self.original_class_or_func.__module__,
            task_namespace=self.task_namespace,
            task_family=self.task_family,
        )

        self.is_class = inspect.isclass(class_or_func)

        # used by decorated UserClass only, stores "wrapped" user class

        # lazy task class definition for orchestration case
        self._task_cls = None  # type: Optional[Type[Task]]
        self._callable_spec = None  # type: Optional[CallableSpec]
        self._callable_tracking_manager = (
            None
        )  # type: Optional[CallableTrackingManager]

    def get_callable_spec(self):
        if not self._callable_spec:
            try:
                self._callable_spec = build_callable_spec(
                    class_or_func=self.original_class_or_func
                )
            except Exception as ex:
                logger.error(
                    "Failed to create task %s: %s\n%s\n",
                    self.original_class_or_func.__name__,
                    str(ex),
                    user_side_code(context=5),
                    exc_info=show_exc_info(ex),
                )
                raise
        return self._callable_spec

    def get_task_cls(self):
        """
        Returns Runnable Task for Orchestration
        """
        if self._task_cls is None:
            # Use the task_type we got from the decorator. check @task/@pipeline/@spark_task
            bases = (self.task_type,)

            self._task_cls = TaskMetaclass(
                str(self.original_class_or_func.__name__),
                bases,
                dict(
                    __doc__=self.original_class_or_func.__doc__,
                    __module__=self.original_class_or_func.__module__,
                    defaults=self.task_defaults,
                    task_decorator=self,
                ),
            )
        return self._task_cls

    def get_task_definition(self):
        return self.get_task_cls().task_definition

    @dbnd_handle_errors(exit_on_error=False)
    def _build_task(self, *args, **kwargs):
        task_cls = self.get_task_cls()
        return task_cls(*args, **kwargs)

    @dbnd_handle_errors(exit_on_error=False)
    def dbnd_run(self, *args, **kwargs):
        # type: (...)-> DatabandRun
        """
        Run task via Databand execution system
        """
        t = self._build_task(*args, **kwargs)
        return t.dbnd_run()

    def tracking_context(self, call_args, call_kwargs):
        if not self._callable_tracking_manager:
            self._callable_tracking_manager = CallableTrackingManager(
                task_decorator=self
            )
        return self._callable_tracking_manager.tracking_context(
            call_args=call_args, call_kwargs=call_kwargs
        )

    def handle_callable_call(self, *call_args, **call_kwargs):
        dbnd_project_config = get_dbnd_project_config()
        if dbnd_project_config.disabled:
            return self.class_or_func(*call_args, **call_kwargs)

        # we are at tracking mode
        if dbnd_project_config.is_tracking_mode():
            with self.tracking_context(call_args, call_kwargs) as track_result_callback:
                fp_result = self.class_or_func(*call_args, **call_kwargs)
                return track_result_callback(fp_result)

        #### DBND ORCHESTRATION MODE
        #
        #     -= Use "Step into My Code"" to get back from dbnd code! =-
        #
        # decorated object call/creation  ( my_func(), MyDecoratedTask()
        # we are at orchestration mode

        task_cls = self.get_task_cls()

        if is_in_airflow_dag_build_context():
            # we are in Airflow DAG building mode - AIP-31
            return build_task_at_airflow_dag_context(
                task_cls=task_cls, call_args=call_args, call_kwargs=call_kwargs
            )

        current = try_get_current_task()
        if not current:
            # no tracking/no orchestration,
            # falling back to "natural call" of the class_or_func
            message = (
                "Can't report tracking info. %s is decorated with @task, but no tracking context was found"
                % (self.class_or_func.__name__,)
            )
            get_one_time_logger().log_once(message, "task_decorator", logging.WARNING)
            return self.class_or_func(*call_args, **call_kwargs)

        ######
        # current is not None, and we are not in tracking/airflow/luigi
        # this is DBND Orchestration mode
        # we can be in the context of task.run() or in task.band()
        # called from user code using user_decorated_func()  or UserDecoratedTask()

        if self.is_class:
            call_kwargs.pop("__call_original_cls", False)

        # we should not get here from _TaskFromTaskDecorator.invoke()
        # at that function we should call user code directly
        phase = current_phase()
        if phase is TaskContextPhase.BUILD:
            # we are in the @pipeline.band() context, we are building execution plan
            t = task_cls(*call_args, **call_kwargs)

            # we are in the band, and if user_code() is called we want to remove redundant
            # `user_code().result` usage
            if t.task_definition.single_result_output:
                return t.result

            # we have multiple outputs (more than one "output" parameter)
            # just return task object, user will use it as `user_code().output_1`
            return t
        elif phase is TaskContextPhase.RUN:
            # we are "running" inside some other task execution (orchestration!)
            #  (inside user_defined_function() or UserDefinedTask.run()

            # if possible we will run it as "orchestration" task
            # with parameters parsing
            if (
                current.settings.run.task_run_at_execution_time_enabled
                and current.task_supports_dynamic_tasks
            ):
                return self._run_task_from_another_task_execution(
                    parent_task=current, call_args=call_args, call_kwargs=call_kwargs
                )
            # we can not call it in "dbnd" way, fallback to normal call
            if self.is_class:
                call_kwargs["__call_original_cls"] = False
            return self.class_or_func(*call_args, **call_kwargs)
        else:
            raise Exception()

    def _run_task_from_another_task_execution(
        self, parent_task, call_args, call_kwargs
    ):
        # type: (TaskDecorator, Task, *Any, **Any) -> TaskRun
        # task is running from another task
        task_cls = self.get_task_cls()
        from dbnd import PipelineTask, pipeline
        from dbnd._core.task_build.dbnd_decorator import _default_output

        dbnd_run = get_databand_run()

        # orig_call_args, orig_call_kwargs = call_args, call_kwargs
        call_args, call_kwargs = args_to_kwargs(
            self.get_callable_spec().args, call_args, call_kwargs
        )

        # Map all kwargs to the "original" target of that objects
        # for example: for DataFrame we'll try to find a relevant target that were used to read it
        # get all possible value's targets
        call_kwargs_as_targets = dbnd_run.target_origin.get_for_map(call_kwargs)
        for p_name, value_origin in call_kwargs_as_targets.items():
            root_target = value_origin.origin_target
            path = root_target.path if hasattr(root_target, "path") else None
            original_object = call_kwargs[p_name]
            call_kwargs[p_name] = InlineTarget(
                root_target=root_target,
                obj=original_object,
                value_type=value_origin.value_type,
                source=value_origin.origin_target.source,
                path=path,
            )

        call_kwargs.setdefault("task_is_dynamic", True)
        call_kwargs.setdefault(
            "task_in_memory_outputs",
            parent_task.settings.run.task_run_at_execution_time_in_memory_outputs,
        )

        if issubclass(task_cls, PipelineTask):
            # if it's pipeline - create new databand run
            # create override _task_default_result to be object instead of target
            task_cls = pipeline(
                self.class_or_func, _task_default_result=_default_output
            ).task_cls

            # instantiate inline pipeline
            task = task_cls(*call_args, **call_kwargs)
            # if it's pipeline - create new databand run
            run = dbnd_run.context.dbnd_run_task(task)
            task_run = run.get_task_run(task.task_id)
        else:
            # instantiate inline task (dbnd object)
            task = task_cls(*call_args, **call_kwargs)

            # update upstream/downstream relations - needed for correct tracking
            # we can have the task as upstream , as it was executed already
            if not parent_task.task_dag.has_upstream(task):
                parent_task.set_upstream(task)

            from dbnd._core.task_build.task_cls__call_state import TaskCallState

            task._dbnd_call_state = TaskCallState(should_store_result=True)
            try:
                task_run = dbnd_run.run_executor.run_task_at_execution_time(
                    task, task_engine=current_task_run().task_engine
                )

                # this will work only for _DecoratedTask
                if task._dbnd_call_state.result_saved:
                    return task._dbnd_call_state.result

            finally:
                # we'd better clean _invoke_result to avoid memory leaks
                task._dbnd_call_state = None

        # if we are inside run, we want to have real values, not deferred!
        if task.task_definition.single_result_output:
            return task.__class__.result.load_from_target(task.result)
            # we have func without result, just fallback to None
        return task

    # compatibility support
    @property
    def task_cls(self):
        return self.get_task_cls()

    @property
    def t(self):
        return self.get_task_cls()

    @property
    def task(self):
        return self.get_task_cls()

    @property
    def task_definition(self):
        return self.get_task_definition()

    @property
    def func(self):
        return self.class_or_func

    @property
    def callable(self):
        return self.class_or_func


class _UserClassWithTaskDecoratorMetaclass(type):
    """
    Used by decorated user classes only,
    1. we change metaclass of original class to go through __call__ on object call
    2. object still behaves as original object (until __call_ is called)
    3. we intercept the call and may call original object, or create dbnd task class
    (at @pipeline or inside @task.run function)

    in order to prevent recursion ( from DecoratedCallableTask.invoke for example)
    we use `__call_original_cls` kwarg. if present we would call an original code

    this code should be serializable with pickle!

    @task
    class UserClass():
        pass
    """

    __is_dbnd_task__ = True

    task_decorator = None  # type: TaskDecorator

    def __call__(cls, *args, **kwargs):
        """
        wrap user class ,so on user_class() we run _item_call first and if required we return task object inplace
        """
        if kwargs.pop("__call_original_cls", False):
            return super(_UserClassWithTaskDecoratorMetaclass, cls).__call__(
                *args, **kwargs
            )

        # prevent recursion call. next time we call cls() we will go into original ctor()
        kwargs["__call_original_cls"] = True
        return cls.task_decorator.handle_callable_call(*args, **kwargs)

    # exposing dbnd logic
    # so OriginalUserClass.task can be used
    # this list should be alligned with attributes at build_dbnd_decorated_func

    @property
    def task_cls(self):
        return self.task_decorator.get_task_cls()

    @property
    def t(self):
        return self.task_cls

    @property
    def task(self):
        return self.task_cls

    @property
    def func(self):
        return self.task_decorator.callable

    @property
    def callable(self):
        return self.task_decorator.callable

    def dbnd_run(self, *args, **kwargs):
        return self.task_decorator.dbnd_run(*args, **kwargs)


def build_dbnd_decorated_func(task_decorator):
    def dbnd_decorated_func(*args, **kwargs):
        """
        DBND Wrapper of User Function
        Redirect call to dbnd logic that might track/orchestrate user code
        """
        return task_decorator.handle_callable_call(*args, **kwargs)

    # new wrapper should should look like original function and be serializable
    # class decorator will not work, because of pickle errors
    functools.update_wrapper(dbnd_decorated_func, task_decorator.original_class_or_func)

    # this list should be alligned with attributes at _UserClassWithTaskDecoratorMetaclass

    # we don't want to create .task_cls object immediately(that is used for orchestration only)
    # however, we can't just return task_decorator, as it's class, and it can not be serialized
    # with pickle, as user func != task_decorator class
    # we wrap "all known" heavy properties with CallableLazyObjectProxy

    lazy_task_cls_property = CallableLazyObjectProxy(task_decorator.get_task_cls)
    # create lazy properties t, task and task_cls
    dbnd_decorated_func.task = lazy_task_cls_property
    dbnd_decorated_func.t = lazy_task_cls_property
    dbnd_decorated_func.task_cls = lazy_task_cls_property

    dbnd_decorated_func.func = task_decorator.callable  # backward compatiblity
    dbnd_decorated_func.callable = task_decorator.callable

    dbnd_decorated_func.dbnd_run = task_decorator.dbnd_run
    dbnd_decorated_func.task_decorator = task_decorator
    dbnd_decorated_func.__is_dbnd_task__ = True
    return dbnd_decorated_func



# © Copyright Databand.ai, an IBM Company 2022

import logging
import sys
import time

from dbnd._core.context.databand_context import new_dbnd_context
from dbnd._core.utils.timeout import timeout
from dbnd._vendor import click


logger = logging.getLogger(__name__)


@click.group(help="Tracker commands to ensure synchronization against webserver")
def tracker():
    pass


@tracker.command(help="Wait for the webserver to initialize")
@click.option(
    "--timeout",
    "-t",
    type=int,
    default=300,
    help="Amount of seconds to wait for webserver until timing out",
)
def wait(timeout):
    with new_dbnd_context(name="new_context") as dbnd_ctx:
        logger.info("Waiting {} seconds for tracker to become ready:".format(timeout))

        is_ready = _wait_for_tracking_store(dbnd_ctx, wait_timeout=timeout)
        if not is_ready:
            logger.error("Tracker is not ready after {} seconds.".format(timeout))
            sys.exit(1)
        logger.info("Tracker is ready.")


def _wait_for_tracking_store(dbnd_ctx, wait_timeout):
    try:
        with timeout(wait_timeout, handler=lambda *args: None):
            start_time = time.time()
            while not dbnd_ctx.tracking_store.is_ready():
                time.sleep(2)  # 2 seconds, 1 doesn't make a difference
                logger.info(
                    "Databand server is still not ready after %s seconds",
                    round(time.time() - start_time),
                )
        return dbnd_ctx.tracking_store.is_ready()
    except TimeoutError:
        pass
    return False



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.utils.string_utils import strip_whitespace
from dbnd._vendor import click
from dbnd._vendor.click.utils import make_default_short_help


logger = logging.getLogger(__name__)


class ConfigValueType(click.ParamType):
    def __init__(self):
        self.name = "config"

    def convert(self, value, param, ctx):
        from dbnd._core.utils import json_utils
        from targets.values.structure import _PARSABLE_PARAM_PREFIX

        value = value.strip()
        # support for --set '{ "a":2, "b":33}'
        if value[0] in _PARSABLE_PARAM_PREFIX:
            return json_utils.loads(value)

        name, sep, var = value.partition("=")
        return {name: var}


def _help(string):
    if not string:
        return "-"
    return strip_whitespace(string)


def _help_short(string):
    return make_default_short_help(string or "-")



# © Copyright Databand.ai, an IBM Company 2022

import json


def build_query_api_params(filters=None, page_number=None, page_size=None, sort=None):
    params_map = [
        ("sort", sort),
        ("filter", json.dumps(filters) if filters else None),
        ("page[number]", page_number),
        ("page[size]", page_size),
    ]
    params = [name + "=" + str(value) for name, value in params_map if value]
    return "&".join(params)


def build_filter(name, operator, value):
    return {"name": name, "op": operator, "val": value}


def create_filter_builder(name, operator):
    def inner_func(value):
        if value is not None:
            return [build_filter(name, operator, value)]
        else:
            return []

    return inner_func


def create_filters_builder(**kwargs):
    """Creates a filters builder using a mapping between filter name and the params relevant for
    `create_filter_builder`.
    Thw new builder function receives mapping between the filter name and the value, creates the filter only if the
    value exists and append it to the filters list"""
    filter_builders = {
        arg_name: create_filter_builder(*filter_params)
        for arg_name, filter_params in kwargs.items()
    }

    def filters_builder(**kwargs):
        filters = []
        for name, value in kwargs.items():
            if name not in filter_builders:
                raise KeyError(
                    "filters_builder has not key argument name {name}".format(name=name)
                )

            filters += filter_builders[name](value)

        return filters

    return filters_builder



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.parameter import PARAMETER_FACTORY


logger = logging.getLogger(__name__)


class ParametersMapper(object):
    def __init__(self):
        self.custom_parameters = {}

    def register_custom_parameter(self, value_type, parameter):
        self.custom_parameters[value_type] = parameter

    def get_parameter(self, value_type):
        custom_parameter = self.custom_parameters.get(value_type)
        if custom_parameter:
            return custom_parameter
        return PARAMETER_FACTORY.modify(value_type=value_type)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import random

from dbnd import log_metric, task
from dbnd._core.current import try_get_databand_context
from dbnd.tasks.doctor.doctor_report_builder import DoctorStatusReportBuilder


logger = logging.getLogger(__name__)


@task
def dbnd_status():
    report = DoctorStatusReportBuilder("Databand Status")

    report.log("env.DBND_HOME", os.environ.get("DBND_HOME"))
    dc = try_get_databand_context()
    report.log("DatabandContext", dc)
    if dc:
        report.log("initialized", dc)

    # calling metrics.
    log_metric("metric_check", "OK")
    log_metric("metric_random_value", random.random())
    return report.get_status_str()



# © Copyright Databand.ai, an IBM Company 2022

import gzip
import json
import logging
import uuid

from datetime import datetime, timedelta
from http import HTTPStatus
from time import sleep
from typing import Dict, Optional, Tuple, Union

import requests

from six.moves.urllib_parse import urljoin

from dbnd._core.current import try_get_databand_run
from dbnd._core.errors.base import (
    DatabandApiError,
    DatabandAuthenticationError,
    DatabandConnectionException,
)
from dbnd._core.errors.friendly_error.api import api_connection_refused
from dbnd._core.log.logging_utils import create_file_handler
from dbnd._core.utils.http.retry_policy import LinearRetryPolicy
from dbnd._vendor import curlify
from dbnd.utils.trace import get_tracing_id


# we'd like to have all requests with default timeout, just in case it's stuck
DEFAULT_REQUEST_TIMEOUT = 300

logger = logging.getLogger(__name__)


# uncomment for requests trace
# import http.client
# http.client.HTTPConnection.debuglevel = 1


class ApiClient(object):
    """Json API client implementation."""

    api_prefix = "/api/v1/"
    default_headers = None

    def __init__(
        self,
        api_base_url: str,
        credentials: Optional[Dict[str, str]] = None,
        debug_server: bool = False,
        session_timeout: int = 5,
        default_max_retry: int = 1,
        default_retry_sleep: Union[int, float] = 0,
        default_request_timeout: Union[
            float, Tuple[float, float]
        ] = DEFAULT_REQUEST_TIMEOUT,
        extra_default_headers: Optional[Dict[str, str]] = None,
        ignore_ssl_errors: bool = False,
    ):
        """
        @param api_base_url: databand webserver url to build the request with
        @param credentials: dict of credential to authenticate with the webserver
         can include "token" key or "username"  and "password" keys
        @param debug_server: flag to debug the webserver - collect logs from webserver and the client's requests
        @param session_timeout: minutes to recreate the requests session
        @param default_max_retry: default value for retries for failed connection
        @param default_retry_sleep: default value for sleep between retries
        @param default_request_timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple
        """

        self._api_base_url = api_base_url

        self.credentials = credentials
        self.default_headers = {
            "Accept": "application/json",
            "content-encoding": "gzip",
            "Content-Type": "application/json",
            **(extra_default_headers or {}),
        }

        self.session: Optional[requests.Session] = None
        self.session_creation_time = None
        self.session_timeout = session_timeout

        self.default_max_retry = default_max_retry
        self.default_retry_sleep = default_retry_sleep
        self.default_request_timeout = default_request_timeout

        self.ignore_ssl_errors = ignore_ssl_errors

        self.debug_mode = debug_server
        if debug_server:
            # the header require so the webserver will record logs and sql queries
            self.default_headers["X-Databand-Debug"] = "True"
            # log the request curl command here
            self.requests_logger = build_file_logger("requests", fmt="%(message)s")
            # log the webserver logs and sql queries here
            self.webserver_logger = build_file_logger("webserver")

    def is_session_expired(self):
        return datetime.now() - self.session_creation_time >= timedelta(
            minutes=self.session_timeout
        )

    def _request(
        self,
        endpoint,
        session,
        method="GET",
        data=None,
        headers=None,
        query=None,
        request_timeout=None,
    ):
        headers = dict(self.default_headers, **(headers or {}))
        url = urljoin(self._api_base_url, endpoint)

        headers["X-Request-ID"] = uuid.uuid4().hex
        headers["X-Databand-Trace-ID"] = get_tracing_id().hex
        request_params = dict(
            method=method,
            url=url,
            data=gzip.compress(json.dumps(data).encode("utf-8")),
            headers=headers,
            params=query,
            timeout=request_timeout or self.default_request_timeout,
            verify=not self.ignore_ssl_errors,
        )
        logger.debug("Sending the following request: %s", request_params)
        resp = self._send_request(session, **request_params)

        if self.debug_mode:
            # save the curl of the current request
            curl_request = curlify.to_curl(resp.request)
            self.requests_logger.info(curl_request)

        if not resp.ok:
            logger.debug("Response is not ok, Raising DatabandApiError")
            if resp.status_code in [HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN]:
                raise DatabandAuthenticationError(
                    "Authentication Error: Failed authenticating to Databand server, please check supplied credentials"
                )

            raise DatabandApiError(
                method, url, resp.status_code, resp.content.decode("utf-8")
            )

        if resp.content:
            try:
                data = resp.json()
            except Exception as e:
                logger.info("Failed to get resp.json(). Exception: {}".format(e))
            else:
                if self.debug_mode and isinstance(data, dict):
                    msg = "api_call: {request}\ndebug info:{debug_data}\n".format(
                        request=resp.request.url, debug_data=data.get("debug")
                    )
                    self.webserver_logger.info(msg)

                return data

        return

    def _send_request(self, session, method, url, **kwargs):
        return session.request(method, url, **kwargs)

    def _authenticate(self):
        if not self.has_credentials():
            return

        try:
            credentials = self.credentials
            token = credentials.get("token")
            if token:
                self.default_headers["Authorization"] = "Bearer {}".format(token)
                return

            if "username" in credentials and "password" in credentials:
                logger.debug("Attempting to login to webserver")
                self.api_request("auth/login", method="POST", data=credentials)
            else:
                logger.warning(
                    "ApiClient._authenticate: username or password is not provided"
                )
        except DatabandConnectionException as e:
            self.remove_session()
            raise e
        except Exception:
            self.remove_session()
            logger.debug("Fail authenticating Databand Webserver ", exc_info=True)
            raise DatabandAuthenticationError(
                "Authentication Error: Failed authenticating to Databand server, please check supplied credentials"
            )

    def create_session(self):
        logger.debug("Initialising session for webserver")
        self.session = requests.session()
        self.session_creation_time = datetime.now()

    def remove_session(self):
        self.session = None
        self.session_creation_time = None

    def api_request(
        self,
        endpoint,
        data,
        method="POST",
        headers=None,
        query=None,
        no_prefix=False,
        retry_policy=None,
        failure_handler=None,
        request_timeout=None,
        requires_auth=True,
    ):
        retry_policy = retry_policy or LinearRetryPolicy(
            seconds_to_sleep=self.default_retry_sleep,
            max_retries=self.default_max_retry,
        )

        url = endpoint if no_prefix else urljoin(self.api_prefix, endpoint)

        retry_number = 0
        while True:
            retry_number += 1
            try:
                if requires_auth:
                    session = self.authenticated_session()
                else:
                    session = self.anonymous_session()

                resp = self._request(
                    url,
                    session=session,
                    method=method,
                    data=data,
                    headers=headers,
                    query=query,
                    request_timeout=request_timeout,
                )
            except (requests.ConnectionError, requests.Timeout) as ex:
                if failure_handler:
                    failure_handler(ex, retry_policy, retry_number)

                if retry_policy.should_retry(500, None, retry_number):
                    sleep(retry_policy.seconds_to_sleep(retry_number))
                    continue

                self.remove_session()
                raise api_connection_refused(self._api_base_url + url, ex)
            return resp

    def is_ready(self):
        try:
            self.api_request(
                "auth/ping",
                None,
                method="GET",
                requires_auth=False,
                retry_policy=LinearRetryPolicy(0, 1),
            )
            return True
        except (DatabandConnectionException, DatabandApiError):
            return False

    def has_credentials(self) -> bool:
        return bool(self.credentials)

    def is_configured(self) -> bool:
        return bool(self._api_base_url)

    def authenticated_session(self):
        if not self.session or self.is_session_expired():
            logger.debug(
                "Webserver session does not exist or timed out, creating new one"
            )
            self.create_session()
            self._authenticate()
        return self.session

    def anonymous_session(self) -> requests.Session:
        return requests.session()

    def __str__(self):
        return "{}({})".format(self.__class__.__name__, self._api_base_url)


def build_file_logger(name, fmt=None):
    """
    Create a logger which write only to a file.
    the file will be located under the run dict.
    """
    file_logger = logging.getLogger("{}_{}".format(__name__, name))
    file_logger.propagate = False

    run = try_get_databand_run()
    if run:
        log_file = run.run_local_root.partition("{}.logs".format(name))
        logger.info(
            "Api-clients {name} logs writing into {path}".format(
                name=name, path=log_file
            )
        )
        handler = create_file_handler(str(log_file), fmt=fmt)
        file_logger.addHandler(handler)
        file_logger.setLevel(logging.INFO)

    return file_logger



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from typing import Iterable, Optional

import six

from dbnd._core.configuration.config_store import _ConfigStore, _lower_config_name
from dbnd._core.utils.basics.text_banner import TextBanner, safe_string, safe_tabulate


if typing.TYPE_CHECKING:
    from dbnd._core.configuration.dbnd_config import DbndConfig

logger = logging.getLogger(__name__)


def pformat_current_config(config, as_table=False, sections=None):
    # type: (DbndConfig, bool, Optional[Iterable[str]]) -> str
    config_layer = config.config_layer
    tb = TextBanner("Config {}".format(config_layer.name))
    tb.column("LAYERS", config_layer.config_layer_path)
    if as_table:
        view_str = pformat_config_store_as_table(
            config_store=config_layer.config, sections=sections
        )
    else:
        value_dict = config_layer.config.as_value_dict(sections=sections)
        view_str = tb.f_struct(value_dict)

    tb.column("CONFIG", view_str)
    return tb.get_banner_str()


def pformat_config_store_as_table(config_store, sections=None):
    # type: (_ConfigStore, Optional[Iterable[str]]) -> str
    header = ["Section", "Key", "Value", "Source", "Priority"]
    data = []
    if sections:
        sections = [_lower_config_name(s) for s in sections]
    else:
        sections = config_store.keys()

    for section in sections:
        section_values = config_store.get(section)
        if not section_values:
            continue
        for key, config_value in six.iteritems(section_values):
            data.append(
                (
                    section,
                    key,
                    safe_string(config_value.value, 300),
                    config_value.source,
                    config_value.priority,
                )
            )

    if data:
        return safe_tabulate(tabular_data=data, headers=header)
    return ""


def pformat_all_layers(config, sections=None):
    # type: (DbndConfig, Optional[Iterable[str]]) -> str
    tb = TextBanner("Configs")

    layers = config.config_layer.get_all_layers()

    # start from the first one
    for layer in reversed(layers):
        layer_values = layer.layer_config.as_value_dict(sections=sections)
        tb.column("LAYER {}".format(layer.config_layer_path), tb.f_struct(layer_values))
    return tb.get_banner_str()



# © Copyright Databand.ai, an IBM Company 2022

from typing import List

from dbnd import Task
from dbnd._core.errors import DatabandError
from dbnd._core.task_ctrl.task_ctrl import TaskSubCtrl


class TaskDescendants(TaskSubCtrl):
    def __init__(self, task):
        super(TaskDescendants, self).__init__(task)

        self.children = set()

    def add_child(self, task_id):
        self.children.add(task_id)

    def get_children(self):
        # type: (...)-> List[Task]
        tic = self.dbnd_context.task_instance_cache
        children = []
        for c_id in self.children:
            child_task = tic.get_task_by_id(c_id)
            if child_task is None:
                raise DatabandError(
                    "You have created %s in different dbnd_context, "
                    "can't find task object in current context!" % c_id
                )
            children.append(child_task)
        return children
